# -*- coding: utf-8 -*-
"""Test the PH image featurizer."""
import pytest

from mofdscribe.featurizers.topology.ph_image import PHImage

from ..helpers import is_jsonable


def test_phimage(hkust_structure, irmof_structure, cof_structure, hkust_la_structure):
    """Ensure we get the correct number of features."""
    phi = PHImage()
    for structure in [hkust_structure, irmof_structure, cof_structure]:
        features = phi.featurize(structure)
        assert len(features) == 20 * 20 * 4 * 3

    assert len(phi.feature_labels()) == 20 * 20 * 4 * 3
    assert is_jsonable(dict(zip(phi.feature_labels(), features)))
    assert features.ndim == 1

    # now, try that encoding chemistry with varying atomic radii works
    # we do this by computing the PH images for structures with same geometry
    # and connectivity, but different atoms
    phi = PHImage(
        atom_types=None, alpha_weight="atomic_radius_calculated", min_size=50, max_b=30, max_p=30
    )
    image_cu = phi.featurize(hkust_structure)
    image_la = phi.featurize(hkust_la_structure)
    assert image_cu.shape == image_la.shape
    assert (image_cu != image_la).any()

    # now, to be sure do not encode the same thing with the atomic radius
    phi = PHImage(atom_types=None, alpha_weight=None, min_size=50, max_b=30, max_p=30)
    image_cu = phi.featurize(hkust_structure)
    image_la = phi.featurize(hkust_la_structure)
    assert image_cu.shape == image_la.shape
    assert image_cu == pytest.approx(image_la, rel=1e-2)


def test_phimage_fit(hkust_structure, irmof_structure):
    """Ensure that calling fit changes the settings."""
    phi = PHImage()
    phi.fit([hkust_structure, irmof_structure])

    assert len(phi.max_b) == len(phi.max_p) == 4
    assert phi.max_b[0] == phi.max_b[3] == 0
    assert 3.0 < phi.max_b[1] < 3.5

    assert 2.6 < phi.max_p[0] < 3.5



# -*- coding: utf-8 -*-
"""Helper functions for obtaining the correct number of replicas"""
from math import ceil, cos, fabs, hypot, pi, sin
from typing import Tuple

import numpy as np
from pymatgen.core import Structure


def resize_unit_cell(structure: Structure, threshold: float) -> Tuple[int, int, int]:
    """Return the multiplication factors for the cell vectors.

    In every direction: min(perpendicular_width) > threshold.

    Args:
        structure (Structure): pymatgen Structure
        threshold (float): threshold for the minimum width of the cell

    Returns:
        Tuple[int, int, int]: Multiplication factors for the cell vectors.
    """
    deg2rad = pi / 180.0

    a_len = structure.lattice.a
    b_len = structure.lattice.b
    c_len = structure.lattice.c

    alpha = structure.lattice.alpha * deg2rad
    beta = structure.lattice.beta * deg2rad
    gamma = structure.lattice.gamma * deg2rad

    # Computing triangular cell matrix
    vol = np.sqrt(
        1
        - cos(alpha) ** 2
        - cos(beta) ** 2
        - cos(gamma) ** 2
        + 2 * cos(alpha) * cos(beta) * cos(gamma)
    )
    cell = np.zeros((3, 3))
    cell[0, :] = [a_len, 0, 0]
    cell[1, :] = [b_len * cos(gamma), b_len * sin(gamma), 0]
    cell[2, :] = [
        c_len * cos(beta),
        c_len * (cos(alpha) - cos(beta) * cos(gamma)) / (sin(gamma)),
        c_len * vol / sin(gamma),
    ]
    cell = np.array(cell)

    # Computing perpendicular widths, as implemented in Raspa
    # for the check (simplified for triangular cell matrix)
    axc1 = cell[0, 0] * cell[2, 2]
    axc2 = -cell[0, 0] * cell[2, 1]
    bxc1 = cell[1, 1] * cell[2, 2]
    bxc2 = -cell[1, 0] * cell[2, 2]
    bxc3 = cell[1, 0] * cell[2, 1] - cell[1, 1] * cell[2, 0]
    det = fabs(cell[0, 0] * cell[1, 1] * cell[2, 2])
    perpwidth = np.zeros(3)
    perpwidth[0] = det / hypot(bxc1, bxc2, bxc3)
    perpwidth[1] = det / hypot(axc1, axc2)
    perpwidth[2] = cell[2, 2]

    # prevent from crashing if threshold value is zero
    thr = max(0.001, threshold)

    return (
        int(ceil(thr / perpwidth[0])),
        int(ceil(thr / perpwidth[1])),
        int(ceil(thr / perpwidth[2])),
    )



# -*- coding: utf-8 -*-
"""
This script does the following

(1) It loops over all the taks it knows. Each tasks corresponds to a subfolder in the top level
`bench` folder.
(2) In each subfolder, it globs for all `.json` file.
(3) For each `.json` file, we look for a corresponding `.rst` file. If there is None, we create it.
(4) We read all `.json` files, validate them, and then compile the results into one long dataframe. We append some metadata for sphinx-needs to the `.rst` file. For each entry, generate a unique id
by taking `R-{first_letter_of_symbol}-{first_letter_of_feat}{year}{month}{day}{hour}{minute}{second}`
(5) This dataframe is used to generate a boxplot which we save as `html`.
(5) We copy the `.rst` and `.html` file to the corresponding `docs/source/leaderboars/` folder.
The `.rst` files will end up in the `docs/source/leaderboards/task_models` folder.
"""

import os
import shutil
from glob import glob
from pathlib import Path

import holoviews as hv
import pandas as pd
from loguru import logger

from mofdscribe.bench.mofbench import BenchResult, BenchTaskEnum

hv.extension("bokeh")


class NoTaskException(Exception):
    pass


METRIC_CARD_TEMPLATE = """Metric card
~~~~~~~~~~~~~~~~~~~

.. regressionmetrics:: {name}
   :id: {id}
   :start_time: {start_time}
   :end_time: {end_time}
   :version: {version}
   :features: {features}
   :name: {name}
   :task: {task}
   :model_type: {model_type}
   :reference: {reference}
   :implementation: {implementation}
   :mofdscribe_version: {mofdscribe_version}
   :mean_squared_error: {mean_squared_error}
   :mean_absolute_error: {mean_absolute_error}
   :r2_score: {r2_score}
   :max_error: {max_error}
   :mean_absolute_percentage_error: {mean_absolute_percentage_error}
   :top_5_in_top_5: {top_5_in_top_5}
   :top_10_in_top_10: {top_10_in_top_10}
   :top_50_in_top_50: {top_50_in_top_50}
   :top_100_in_top_100: {top_100_in_top_100}
   :top_500_in_top_500: {top_500_in_top_500}
   :session_info: {session_info}
"""

DOC_DIR = "../docs"


def make_plot(df, outname):
    cols = [
        "mean squared error",
        "mean absolute error",
        "r2 score",
        "max error",
        "mean absolute percentage error",
        "top 5 in top 5",
        "top 10 in top 10",
        "top 50 in top 50",
        "top 100 in top 100",
        "top 500 in top 500",
    ]

    f = hv.HoloMap(
        {column: hv.BoxWhisker(df, kdims="name", vdims=column) for column in cols}, kdims="metric"
    ).opts(framewise=True, width=450, invert_axes=True)
    hv.util.output(
        widget_location="right",
    )
    hv.save(f, outname, fmt="html")


def update_rst(file, bench_result):
    mean_metrics = bench_result.metrics.average_metrics()
    id = Path(file).stem
    top_keys = [
        "name",
        "start_time",
        "end_time",
        "version",
        "features",
        "name",
        "task",
        "model_type",
        "reference",
        "mofdscribe_version",
        "implementation",
        "session_info",
    ]
    card_dict = {}
    for key in top_keys:
        card_dict[key] = getattr(bench_result, key)

    for k, v in dict(mean_metrics).items():
        try:
            card_dict[k] = v.round(2)
        except AttributeError:
            # integers
            card_dict[k] = v
    card_dict["id"] = id

    with open(file, "r") as handle:
        content = handle.read()

    content += "\n\n" + METRIC_CARD_TEMPLATE.format(**card_dict)

    stem = Path(file).stem
    file_new = file.replace(stem, "R-" + stem)
    with open(file_new, "w") as handle:
        handle.write(content)

    return file_new


def compile_task(task):
    """
    Compile the results of a task into a dataframe.
    """
    # glob for all json files
    task_dir = os.path.join("..", "bench_results", str(task))
    logger.info("Compiling task {}".format(task))

    versions = os.listdir(task_dir)
    latest_version = max(versions)
    for version in versions:
        versionname = "latest" if version == latest_version else version
        json_files = glob(os.path.join(task_dir, version, "*.json"))
        if len(json_files) == 0:
            raise NoTaskException()
        logger.info(f"Found {len(json_files)} json files")
        stems = [Path(f).stem for f in json_files]
        # find all matching rst file
        rst_files = [os.path.join(task_dir, version, f + ".rst") for f in stems]
        # check if the .rst files exist, otherwise create
        for rst_file in rst_files:
            if not os.path.exists(rst_file):
                with open(rst_file, "w") as f:
                    f.write("")

        # read all json files
        dfs = []

        for json in json_files:
            res = BenchResult.parse_file(json)

            df = pd.DataFrame(res.metrics.concatenated_metrics().dict())
            df["name"] = res.name

            dfs.append(df)

        df_all = pd.concat(dfs)

        df_all.columns = [c.replace("_", " ") for c in df_all.columns]
        df_all.to_csv(os.path.join(task_dir, version, "metrics.csv"))
        html_path = os.path.join(task_dir, version, f"{task}_plot_{versionname}.html")
        make_plot(df_all, html_path)

        for rst_file, json in zip(rst_files, json_files):
            new_file = update_rst(rst_file, BenchResult.parse_file(json))
            shutil.copy(
                new_file,
                os.path.join(DOC_DIR, "source", "leaderboards", f"{task}_models", versionname),
            )

        shutil.copy(html_path, os.path.join(DOC_DIR, "source", "leaderboards"))


if __name__ == "__main__":
    for task in BenchTaskEnum._member_names_:
        try:
            compile_task(task)
        except NoTaskException:
            pass



# -*- coding: utf-8 -*-
"""Helper functions for tests."""
import json


def is_jsonable(x):
    """Test if a object is JSONable."""
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False



# -*- coding: utf-8 -*-
"""Test QMOF dataset."""
import numpy as np
import pytest
from pymatgen.core import IStructure

from mofdscribe.datasets.qmof_dataset import QMOFDataset


@pytest.mark.xdist_group(name="qmof-ds")
@pytest.mark.parametrize("flavor", ["all", "csd", "gcmc", "csd-gcmc"])
def test_qmof(flavor):
    """Ensure we can instantiate the QMOF dataset and access a few key methods."""
    qmof = QMOFDataset(flavor=flavor, drop_nan=False)
    assert isinstance(list(qmof.get_structures([1]))[0], IStructure)
    assert isinstance(qmof.get_labels([1]), np.ndarray)
    assert isinstance(qmof.get_years([1]), np.ndarray)
    assert len(qmof.get_years([1])) == 1
    assert len(qmof.get_years([1, 2, 4])) == 3
    assert isinstance(qmof.get_labels([1], ["outputs.pbe.bandgap"]), np.ndarray)

    # make sure we can get a subset of the dataset
    qmof_subset = qmof.get_subset([1, 2, 3, 8])
    assert isinstance(qmof_subset, QMOFDataset)
    assert len(qmof_subset._df) == 4
    assert (
        qmof._df.iloc[[1, 2, 3, 8]]["info.qmof_id"].values == qmof_subset._df["info.qmof_id"].values
    ).all()
    assert list(qmof.get_structures([1]))[0] == list(qmof_subset.get_structures([0]))[0]

    if flavor == "all":
        assert len(qmof) > 10_000



# -*- coding: utf-8 -*-
"""Helper functions for the host-guest featurizers."""
from typing import List, Optional, Union

from pydantic import BaseModel
from pymatgen.core import IMolecule, IStructure, Molecule, Structure
from structuregraph_helpers.subgraph import get_subgraphs_as_molecules

from mofdscribe.featurizers.bu.utils import boxed_molecule
from mofdscribe.featurizers.utils.structure_graph import get_structure_graph


def remove_guests_from_structure(structure, guest_indices: List[List[int]]) -> Structure:
    """
    Remove guests from a structure.

    Args:
        structure (Structure): The structure to remove guests from.
        guest_indices (List[List[int]]): The indices of the guests.

    Returns:
        Structure: The structure without guests.
    """
    flattened_guest_indices = sum(guest_indices, [])
    ok_sites = []
    for i, site in enumerate(structure):
        if i not in flattened_guest_indices:
            ok_sites.append(site)

    return Structure.from_sites(ok_sites)


class HostGuest(BaseModel):
    """Container for host and guests."""

    host: Union[Structure, IStructure]
    guests: Optional[List[Union[Structure, Molecule, IStructure, IMolecule]]]


def _extract_host_guest(
    structure: Optional[Union[Structure, IStructure]] = None,
    host_guest: Optional[HostGuest] = None,
    operates_on: str = "structure",
    remove_guests: bool = True,
    local_env_method: str = "vesta",
):
    if structure is None and host_guest is None:
        raise ValueError("You must provide a structure or host_guest.")

    if structure is not None:
        if not isinstance(structure, (IStructure)):
            structure = IStructure.from_sites(structure.sites)
        structure_graph = get_structure_graph(structure, local_env_method)
        mols, _mol_graphs, mol_indices, _centers, _coordinates = get_subgraphs_as_molecules(
            structure_graph
        )
        if remove_guests:
            host = remove_guests_from_structure(structure_graph.structure, mol_indices)
        else:
            host = structure_graph.structure

        if operates_on == "structure":
            mols = [boxed_molecule(mol) for mol in mols]

    else:
        host = host_guest.host
        mols = host_guest.guests
        if operates_on == "structure" and isinstance(mols[0], (Molecule, IMolecule)):
            mols = [boxed_molecule(mol) for mol in mols]

    return HostGuest(host=host, guests=mols)



# -*- coding: utf-8 -*-
"""Test helper functions for dealing with structure graphs."""
from mofdscribe.featurizers.utils.structure_graph import get_neighbors_at_distance


def test_get_neighbors_at_distance(hkust_structure_graph):
    """Test the get_neighbors_at_distance function
    at some case studies (manually verified)."""
    assert len(get_neighbors_at_distance(hkust_structure_graph, 1, 1)[1]) == 4

    assert len(get_neighbors_at_distance(hkust_structure_graph, 1, 2)[1]) == 4



# -*- coding: utf-8 -*-
"""Test the functions in the regression metrics module."""
import numpy as np

from mofdscribe.metrics.regression import RegressionMetrics, get_regression_metrics, top_n_in_top_k


def test_top_n_in_top_k():
    """Test the identity of indiscernibles property for top_n_in_top_k."""
    true = np.arange(10)
    pred = np.arange(10)

    assert top_n_in_top_k(pred, true, k=1, n=1) == 1
    assert top_n_in_top_k(pred, true, k=1, n=2) == 1
    assert top_n_in_top_k(pred, true, k=2, n=2) == 2

    pred = pred[::-1]
    assert top_n_in_top_k(pred, true, k=1, n=1) == 0


def test_get_regression_metrics():
    """Test call to get_regression_metrics.

    * Test dentity of indiscernibles property
    * Test return type is RegressionMetrics
    """
    true = np.arange(10)
    pred = np.arange(10)

    res = get_regression_metrics(pred, true)
    assert isinstance(res, RegressionMetrics)
    for k, v in res.dict().items():
        if "top" not in k:
            assert isinstance(v, float)
        else:
            assert isinstance(v, int)
        if k == "r2_score":
            assert v == 1.0
        elif "top" in k:
            assert v >= 1
        else:
            assert v == 0



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import re
import sys
from datetime import date

sys.path.insert(0, os.path.abspath("../../src"))

# -- Project information -----------------------------------------------------

project = "mofdscribe"
copyright = f"{date.today().year}, Kevin M. Jablonka"
author = "Kevin M. Jablonka"

# The full version, including alpha/beta/rc tags.
release = "0.0.7"

# The short X.Y version.
parsed_version = re.match(
    r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?:-(?P<release>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?(?:\+(?P<build>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?",
    release,
)
version = parsed_version.expand(r"\g<major>.\g<minor>.\g<patch>")

if parsed_version.group("release"):
    tags.add("prerelease")

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# A list of prefixes that are ignored when creating the module index. (new in Sphinx 0.6)
modindex_common_prefix = ["mofdscribe."]

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.katex",
    "sphinx_copybutton",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.needs",
    "sphinx.ext.intersphinx",
    "sphinx-pydantic",
]

intersphinx_mapping = {
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}

needs_extra_options = [
    "start_time",
    "end_time",
    "version",
    "features",
    "name",
    "task",
    "model_type",
    "reference",
    "implementation",
    "mofdscribe_version",
    "mean_squared_error",
    "mean_absolute_error",
    "r2_score",
    "max_error",
    "mean_absolute_percentage_error",
    "top_5_in_top_5",
    "top_10_in_top_10",
    "top_50_in_top_50",
    "top_100_in_top_100",
    "top_500_in_top_500",
    "references",
    "considers_geometry",
    "considers_structure_graph",
    "encodes_chemistry",
    "scalar",
    "scope",
    "session_info",
]

needs_types = [
    dict(
        directive="regressionmetrics",
        title="Regression Metrics",
        prefix="R_",
        color="#DCB239",
        style="node",
    ),
    dict(
        directive="featurizer",
        title="Featurizers",
        prefix="F_",
        color="#FFB039",
        style="node",
    ),
]

needs_id_regex = "[A-Za-z0-9 .():_]+"

autodoc_type_aliases = {"ArrayLike": "numpy.typing.ArrayLike"}

copybutton_selector = "div:not(.no-copy)>div.highlight pre"
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# generate autosummary pages
autosummary_generate = True


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


needs_table_style = "datatables"
# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "globaltoc_collapse": False,
    "source_repository": "https://github.com/kjappelbaum/mofdscribe/",
    "source_branch": "main",
    "source_directory": "docs/source",
    # Since Furo doesn't allow us to disable dark mode, we make dark mode
    # equivalent to light mode by overriding all colors back to their light value.
    # See: https://github.com/pradyunsg/furo/issues/28
    "dark_css_variables": {
        # Taken from: https://github.com/pradyunsg/furo/blob/c682d5d3502f3fa713c909eebbf9f3afa0f469d9/src/furo/assets/styles/variables/_colors.scss
        "color-problematic": "#b30000",
        # Base Colors
        "color-foreground-primary": "black",  # for main text and headings
        "color-foreground-secondary": "#5a5c63",  # for secondary text
        "color-foreground-muted": "#646776",  # for muted text
        "color-foreground-border": "#878787",  # for content borders
        "color-background-primary": "white",  # for content
        "color-background-secondary": "#f8f9fb",  # for navigation + ToC
        "color-background-hover": "#efeff4ff",  # for navigation-item hover
        "color-background-hover--transparent": "#efeff400",
        "color-background-border": "#eeebee",  # for UI borders
        "color-background-item": "#ccc",  # for "background" items (eg: copybutton)
        # Announcements
        "color-announcement-background": "#000000dd",
        "color-announcement-text": "#eeebee",
        # Brand colors
        "color-brand-primary": "#2962ff",
        "color-brand-content": "#2a5adf",
        # Highlighted text (search)
        "color-highlighted-background": "#ddeeff",
        # GUI Labels
        "color-guilabel-background": "#ddeeff80",
        "color-guilabel-border": "#bedaf580",
        # API documentation
        "color-api-keyword": "var(--color-foreground-secondary)",
        "color-highlight-on-target": "#ffffcc",
        # Admonitions
        "color-admonition-background": "transparent",
        # Cards
        "color-card-border": "var(--color-background-secondary)",
        "color-card-background": "transparent",
        "color-card-marginals-background": "var(--color-background-hover)",
        # Code blocks
        "color-code-foreground": "black",
        "color-code-background": "#f8f9fb",
        "bk-input-color": "white",
    },
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#
if os.path.exists("figures/logo.png"):
    html_logo = "figures/logo.png"

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "mofdscribedoc"

# -- Options for LaTeX output ------------------------------------------------

# latex_elements = {
#     The paper size ('letterpaper' or 'a4paper').
#
#     'papersize': 'letterpaper',
#
#     The font size ('10pt', '11pt' or '12pt').
#
#     'pointsize': '10pt',
#
#     Additional stuff for the LaTeX preamble.
#
#     'preamble': '',
#
#     Latex figure (float) alignment
#
#     'figure_align': 'htbp',
# }

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
# latex_documents = [
#     (
#         master_doc,
#         'mofdscribe.tex',
#         'mofdscribe Documentation',
#         author,
#         'manual',
#     ),
# ]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "mofdscribe",
        "mofdscribe Documentation",
        [author],
        1,
    ),
]

# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "mofdscribe",
        "mofdscribe Documentation",
        author,
        "Kevin M. Jablonka",
        "Ecosystem for digital reticular chemistry",
        "Miscellaneous",
    ),
]

# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
# epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
# epub_exclude_files = ['search.html']

# -- Extension configuration -------------------------------------------------


autoclass_content = "both"

# Don't sort alphabetically, explained at:
# https://stackoverflow.com/questions/37209921/python-how-not-to-sort-sphinx-output-in-alphabetical-order
autodoc_member_order = "bysource"

html_css_files = [
    "css/overrides.css",
]



# -*- coding: utf-8 -*-
"""Run eqeq."""
import contextlib
import os
from functools import lru_cache
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import List, Tuple

from ase.io.cif import write_cif
from pymatgen.core import IStructure
from pymatgen.io.ase import AseAtomsAdaptor


@lru_cache(maxsize=32)
def get_eqeq_charges(structure: IStructure) -> Tuple[str, List[float]]:
    from pyeqeq import run_on_cif

    with open(os.devnull, "w") as devnull, contextlib.redirect_stdout(devnull), NamedTemporaryFile(
        "w", suffix=".cif"
    ) as f:
        structure.to(filename=f.name, fmt="cif")
        charges = run_on_cif(f.name)

    atoms = AseAtomsAdaptor().get_atoms(structure)
    charge_dict = dict(zip(range(len(atoms)), charges))
    d = {"atom_site_charge": {0: charge_dict}}

    bo = BytesIO()

    write_cif(bo, atoms, loop_keys=d)

    return bo.getvalue().decode("utf-8"), charges



# -*- coding: utf-8 -*-
"""Naive fragmentation implementation following Moosavi et al.

This does not require multiple loops over the graph.
For alternative (slower) implementation see MOFfragmentor.
"""


from copy import deepcopy
from typing import Dict, List, Set

from pymatgen.analysis.graphs import StructureGraph
from pymatgen.core import Structure
from structuregraph_helpers.subgraph import get_subgraphs_as_molecules

from mofdscribe.featurizers.utils.structure_graph import get_connected_site_indices
from mofdscribe.featurizers.utils.substructures import (
    _not_relevant_structure_indices,
    get_metal_indices,
)


def get_node_atoms(structure_graph: StructureGraph) -> Set[int]:
    """Get the indices of the atoms that are assigned as node.

    We do this by identifying the metals and their connected atoms.

    Args:
        structure_graph (StructureGraph): pymatgen StructureGraph

    Returns:
        Set[int]: node indices
    """
    metal_indices = get_metal_indices(structure_graph.structure)
    metal_names = [str(structure_graph.structure[i].specie.symbol) for i in metal_indices]

    node_set = set()
    node_set.update(metal_indices)

    tmp_node_set = set()
    for metal_index in metal_indices:
        tmp_node_set.add(metal_index)
        bonded_to_metal = get_connected_site_indices(structure_graph, metal_index)
        tmp_node_set.update(bonded_to_metal)

    # add atoms that are only connected to metal or hydrogen to the node list + hydrogen atoms connected to them
    for node_atom_index in tmp_node_set:
        all_bonded_atoms = get_connected_site_indices(structure_graph, node_atom_index)
        only_bonded_metal_hydrogen = True
        for index in all_bonded_atoms:
            if not ((str(structure_graph.structure[index].specie) == "H") or index in metal_names):
                only_bonded_metal_hydrogen = False
        if only_bonded_metal_hydrogen:
            node_set.add(node_atom_index)

    final_node_atom_set = deepcopy(node_set)
    for atom_index in node_set:
        for index in get_connected_site_indices(structure_graph, atom_index):
            if str(structure_graph.structure[index].specie) == "H":
                final_node_atom_set.add(index)

    return final_node_atom_set


def get_floating_indices(structure_graph: StructureGraph) -> Set[int]:
    """Get the indices of floating (solvent) molecules in the structure."""
    _, _, idx, _, _ = get_subgraphs_as_molecules(
        structure_graph,
        disable_boundary_crossing_check=False,
        filter_in_cell=True,
        return_unique=True,
    )
    return set(sum(idx, []))


def get_bbs_from_indices(structure_graph: StructureGraph, indices: Set[int]):
    graph_ = structure_graph.__copy__()
    graph_.structure = Structure.from_sites(graph_.structure.sites)
    to_delete = _not_relevant_structure_indices(graph_.structure, indices)
    assert len(structure_graph) == len(to_delete) + len(indices)
    graph_.remove_nodes(to_delete)
    assert len(graph_) == len(structure_graph) - len(to_delete)
    mol, return_subgraphs, idx, centers, coordinates = get_subgraphs_as_molecules(
        graph_,
        return_unique=False,
        disable_boundary_crossing_check=False,
        filter_in_cell=True,
    )
    return mol, return_subgraphs, idx, centers, coordinates


def get_bb_indices(structure_graph: StructureGraph) -> Dict[str, List[List[int]]]:
    node_atoms = get_node_atoms(structure_graph)
    floating_indices = get_floating_indices(structure_graph)
    all_atoms = set(list(range(len(structure_graph))))
    linker_atoms = all_atoms - node_atoms - floating_indices
    (
        linker_mol,
        linker_subgraph,
        linker_idx,
        linker_center,
        linker_coordinates,
    ) = get_bbs_from_indices(structure_graph, linker_atoms)
    node_mol, node_subgraph, node_idx, node_center, node_coordinates = get_bbs_from_indices(
        structure_graph, node_atoms
    )

    linker_atom_types = _linker_atom_types(linker_idx, node_atoms, structure_graph)

    linker_atom_types["nodes"] = node_idx

    return linker_atom_types


def _filter_out_empty_lists(list_of_lists):
    return [x for x in list_of_lists if x]


def _linker_atom_types(indices, node_indices, structure_graph):
    """Group linker atoms in `connecting`, `functional_group` and `scaffold`."""
    functional_group = []
    scaffold = []
    connecting = []

    for index_group in indices:
        functional_group_ = []
        scaffold_ = []
        connecting_ = []
        for index in index_group:
            neighors = get_connected_site_indices(structure_graph, index)
            if any(i in node_indices for i in neighors):
                connecting_.append(index)
            elif structure_graph.structure[index].specie.symbol not in ("H", "C"):
                functional_group_.append(index)
            else:
                scaffold_.append(index)
        functional_group.append(functional_group_)
        scaffold.append(scaffold_)
        connecting.append(connecting_)

    return {
        "linker_all": _filter_out_empty_lists(indices),
        "linker_functional": _filter_out_empty_lists(functional_group),
        "linker_scaffold": _filter_out_empty_lists(scaffold),
        "linker_connecting": _filter_out_empty_lists(connecting),
    }



# -*- coding: utf-8 -*-
"""Partial charge histogram featurizer."""
from typing import List, Union

import numpy as np
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils.eqeq import get_eqeq_charges
from mofdscribe.featurizers.utils.extend import operates_on_istructure, operates_on_structure
from mofdscribe.featurizers.utils.histogram import get_rdf

__all__ = ["PartialChargeHistogram"]


@operates_on_istructure
@operates_on_structure
class PartialChargeHistogram(MOFBaseFeaturizer):
    """Compute partial charges using the EqEq charge equilibration method [Ongari2019]_.

    Then derive a fix-length feature vector from the partial charges by binning
    charges in a histogram.
    """

    def __init__(
        self,
        min_charge: float = -4,
        max_charge: float = 4,
        bin_size: float = 0.5,
        primitive: bool = False,
    ) -> None:
        """Construct a new PartialChargeHistogram featurizer.

        Args:
            min_charge (float): Minimum limit of bin grid.
                Defaults to -4.
            max_charge (float): Maximum limit of bin grid.
                Defaults to 4.
            bin_size (float): Bin size.
                Defaults to 0.5.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        self.min_charge = min_charge
        self.max_charge = max_charge
        self.bin_size = bin_size
        super().__init__(primitive=primitive)

    def _get_grid(self):
        return np.arange(self.min_charge, self.max_charge, self.bin_size)

    def feature_labels(self) -> List[str]:
        return [f"chargehist_{val}" for val in self._get_grid()]

    def _featurize(self, s: Union[Structure, IStructure]) -> np.ndarray:
        if isinstance(s, Structure):
            s = IStructure.from_sites(s.sites)
        _, results = get_eqeq_charges(s)

        hist = get_rdf(results, self.min_charge, self.max_charge, self.bin_size, None, None, False)
        return hist

    def citations(self) -> List[str]:
        return [
            "@article{Ongari2018,"
            "doi = {10.1021/acs.jctc.8b00669},"
            "url = {https://doi.org/10.1021/acs.jctc.8b00669},"
            "year = {2018},"
            "month = nov,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {15},"
            "number = {1},"
            "pages = {382--401},"
            "author = {Daniele Ongari and Peter G. Boyd and Ozge Kadioglu and "
            "Amber K. Mace and Seda Keskin and Berend Smit},"
            "title = {Evaluating Charge Equilibration Methods To Generate "
            "Electrostatic Fields in Nanoporous Materials},"
            "journal = {Journal of Chemical Theory and Computation}"
            "}",
            "@article{Wilmer2012,"
            "doi = {10.1021/jz3008485},"
            "url = {https://doi.org/10.1021/jz3008485},"
            "year = {2012},"
            "month = aug,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {3},"
            "number = {17},"
            "pages = {2506--2511},"
            "author = {Christopher E. Wilmer and Ki Chul Kim and Randall Q. Snurr},"
            "title = {An Extended Charge Equilibration Method},"
            "journal = {The Journal of Physical Chemistry Letters}"
            "}",
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka", "Daniele Ongari", "Christopher Wilmer"]



# -*- coding: utf-8 -*-
"""Test the adversial validation"""
import numpy as np
import pytest

from mofdscribe.datasets import CoREDataset
from mofdscribe.metrics.adverserial import AdverserialValidator
from mofdscribe.splitters import DensitySplitter, HashSplitter


@pytest.mark.xdist_group(name="core-ds")
def test_adverserial_validator():
    """Test the adverserial validation"""
    dataset = CoREDataset()
    splitter = HashSplitter(dataset, sample_frac=0.2)

    feature_names = list(dataset.available_features)

    train_idx, test_idx = splitter.train_test_split(frac_train=0.5)

    x_a = dataset._df[feature_names].iloc[train_idx]
    x_b = dataset._df[feature_names].iloc[test_idx]

    adv = AdverserialValidator(x_a, x_b)
    score = adv.score()
    assert len(score) == 5
    assert np.abs(score.mean() - 0.5) < 0.25


@pytest.mark.xdist_group(name="core-ds")
def test_adverserial_validator_with_different_dist():
    """We use the DensitySplitter to create different distributions.

    Make sure that we detect this and highlight reasonable features.
    """
    dataset = CoREDataset()
    splitter = DensitySplitter(dataset, sample_frac=0.5, density_q=[0, 0.5, 1])

    densities = dataset.get_densities(range(len(dataset)))
    groups = splitter._get_groups()

    assert (
        np.abs(densities[groups == 0].mean() - densities[groups == 1].mean())
        > densities[groups == 0].std()
    )

    feature_names = list(dataset.available_features)

    train_idx, test_idx = splitter.train_test_split(frac_train=0.5)

    x_a = dataset._df[feature_names].iloc[train_idx]
    x_b = dataset._df[feature_names].iloc[test_idx]

    assert (
        np.abs(densities[train_idx].mean() - densities[test_idx].mean())
        > densities[train_idx].std()
    )
    adv = AdverserialValidator(x_a, x_b)
    score = adv.score()
    assert len(score) == 5
    assert np.abs(score.mean() - 0.5) > 0.3



# -*- coding: utf-8 -*-
"""Simple baseline models."""
from typing import Dict, Optional

import numpy as np
from pymatgen.core import Structure
from sklearn.dummy import DummyRegressor as SklearnDummyRegressor
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import PolynomialFeatures, StandardScaler

__all__ = ["DensityRegressor"]


class DensityRegressor:
    """Dummy model."""

    def __init__(self, lr_kwargs: Optional[Dict] = None):
        """Initialize the model.

        Args:
            lr_kwargs (Optional[Dict], optional): Keyword arguments
                that are passed to the linear regressor.
                Defaults to None.
        """
        self.model = Pipeline(
            [
                ("scaler", StandardScaler()),
                ("poly", PolynomialFeatures(degree=3)),
                ("lr", LinearRegression(**(lr_kwargs or {}))),
            ]
        )

    def featurize(self, s: Structure) -> float:
        """You might want to use a lookup in some dataframe instead.

        Or use some mofdscribe featurizers.

        Args:
            s (Structure): Structure to featurize.

        Returns:
            float: Density of the structure.
        """
        return s.density

    def fit(self, idx, structures, y):
        x = np.array([self.featurize(s) for s in structures]).reshape(-1, 1)
        self.model.fit(x, y)

    def predict(self, idx, structures):
        x = np.array([self.featurize(s) for s in structures]).reshape(-1, 1)
        return self.model.predict(x)


class DummyRegressor:
    """Dummy regressor model."""

    def __init__(self, strategy: str = "mean"):
        """Initialize the model.

        Args:
            strategy (str): Strategy to use for prediction.
                Defaults to "mean".
        """
        self.model = SklearnDummyRegressor(strategy=strategy)

    def fit(self, idx, structures, y):
        self.model.fit(idx, y)

    def predict(self, idx, structures):
        return self.model.predict(idx)



# -*- coding: utf-8 -*-
"""Convenience functions for mofdscribe."""

import sys
from typing import List

from loguru import logger

__all__ = ["enable_logging"]


def enable_logging() -> List[int]:
    """Set up the mofdscribe logging with sane defaults."""
    logger.enable("mofdscribe")

    config = dict(
        handlers=[
            dict(
                sink=sys.stderr,
                format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS Z UTC}</>"
                " <red>|</> <lvl>{level}</> <red>|</> <cyan>{name}:{function}:{line}</>"
                " <red>|</> <lvl>{message}</>",
                level="INFO",
            ),
            dict(
                sink=sys.stderr,
                format="<red>{time:YYYY-MM-DD HH:mm:ss.SSS Z UTC} | {level} | {name}:{function}:{line} | {message}</>",
                level="WARNING",
            ),
        ]
    )
    return logger.configure(**config)



# -*- coding: utf-8 -*-
from mofdscribe.featurizers.bu.compositionstats_featurizer import CompositionStats


def test_composition_stats_featurizer(molecule, linker_molecule, triangle_structure):
    """Test the composition stats featurizer."""
    for mol in (molecule, linker_molecule, triangle_structure):
        featurizer = CompositionStats()
        feats = featurizer.featurize(mol)
        assert len(feats) == 8
        for f in feats:
            assert f >= 0
        assert len(feats) == len(featurizer.feature_labels())



# -*- coding: utf-8 -*-
"""Implements persistent homology images"""
from collections import defaultdict
from typing import List, Optional, Tuple, Union

import numpy as np
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils.extend import (
    operates_on_imolecule,
    operates_on_istructure,
    operates_on_molecule,
    operates_on_structure,
)

from ._tda_helpers import (
    get_persistence_image_limits_for_structure,
    get_persistent_images_for_structure,
)


@operates_on_imolecule
@operates_on_molecule
@operates_on_istructure
@operates_on_structure
class PHImage(MOFBaseFeaturizer):
    r"""Vectorize persistent diagrams as image.

    `Adams et al. (2017)
    <https://www.jmlr.org/papers/volume18/16-337/16-337.pdf>`_ introduced a
    stable vector representation of persistent homology.

    In persistent images, one replaces birth–persistence pairs (b, d – b) by a
    Gaussians (to spread its influence across the neighborhood, since nearby
    points represent features of similar size). Additionally, one multiplies
    with special weighting function such that

    .. math::

        f(x, y)=w(y) \sum_{(b, d) \in D_{g}(P)} N((b, d-b), \sigma)

    A common weighting function is the linear function :math:`w(y) = y`.

    One application for porous materials has been reported by `Aditi S.
    Krishnapriyan et al. (2017)
    <https://pubs.acs.org/doi/full/10.1021/acs.jpcc.0c01167>`_.

    Typically, persistent images are computed for all atoms in the structure.
    However, one can also compute persistent images for a subset of atoms. This
    can be done by specifying the atom types in the constructor.
    """

    def __init__(
        self,
        atom_types: Optional[Tuple[str]] = (
            "C-H-N-O",
            "F-Cl-Br-I",
            "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V"
            "-Ag-Nd-U-Ba-Ce-K-Ga-Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-"
            "Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-Hf-Ir-Nb-Pd-Hg-"
            "Th-Np-Lu-Rh-Pu",
        ),
        dimensions: Tuple[int] = (0, 1, 2),
        compute_for_all_elements: bool = True,
        min_size: int = 20,
        image_size: Tuple[int] = (20, 20),
        spread: float = 0.2,
        weight: str = "identity",
        max_b: Union[int, List[int]] = 18,
        max_p: Union[int, List[int]] = 18,
        max_fit_tolerence: float = 0.1,
        periodic: bool = False,
        no_supercell: bool = False,
        primitive: bool = False,
        alpha_weight: Optional[str] = None,
    ) -> None:
        """Construct a PHImage object.

        Args:
            atom_types (Tuple[str], optional): Atoms that are used to create
                substructures that are analysed using persistent homology.
                If multiple atom types separated by hash are provided, e.g.
                "C-H-N-O", then the substructure consists of all atoms of type
                C, H, N, or O. Defaults to ( "C-H-N-O", "F-Cl-Br-I",
                "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V-Ag-Nd-U-Ba-
                Ce-K-Ga-Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-Ho-Re-Be-Rb-La-Sn-Cs-Pb-
                Pr-Bi-Tm-Sr-Ti-Hf-Ir-Nb-Pd-Hg-Th-Np-Lu-Rh-Pu").
            dimensions (Tuple[int]): Dimensions of topological
                features to consider for persistence images. Defaults to (0, 1, 2).
            compute_for_all_elements (bool): If true, compute
                persistence images for full structure (i.e. with all elements). If
                false, it will only do it for the substructures specified with
                `atom_types`. Defaults to True.
            min_size (int): Minimum supercell size (in Angstrom).
                Defaults to 20.
            image_size (Tuple[int]): Size of persistent image in pixel.
                Defaults to (20, 20).
            spread (float): "Smearing factor" for the
                Gaussians. Defaults to 0.2.
            weight (str): Weighting function for calculation of
                the persistence images. Defaults to "identity".
            max_b (Union[int, List[int]]): Maximum birth time.
                Defaults to 18.
            max_p (Union[int, List[int]]): Maximum
                persistence. Defaults to 18.
            max_fit_tolerence (float): If
                `fit` method is used to find the limits of the persistent images,
                one can appy a tolerance on the the found limits. The maximum
                will then be max + max_fit_tolerance * max. Defaults to 0.1.
            periodic (bool): If true, then periodic Euclidean is used
                in the analysis (experimental!). Defaults to False.
            no_supercell (bool): If true, then the supercell is not created.
                Defaults to False.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to False.
            alpha_weight (Optional[str]): If specified, the use weighted alpha shapes,
                i.e., replacing the points with balls of varying radii.
                For instance `atomic_radius_calculated` or `van_der_waals_radius`.

        Raises:
            AssertionError: If the length of the max_b and max_p is not equal
                to the number of dimensions.
        """
        atom_types = [] if atom_types is None else atom_types
        self.atom_types = atom_types
        self.compute_for_all_elements = compute_for_all_elements
        self.min_size = min_size
        self.image_size = image_size
        self.spread = spread
        self.dimensions = dimensions
        self.weight = weight
        if isinstance(max_b, (list, tuple)):
            if len(max_b) != len(dimensions):
                raise AssertionError(
                    "max_b must be a list of length equal to the number of dimensions"
                )
        else:
            max_b = [max_b] * len(dimensions)

        if isinstance(max_p, (list, tuple)):
            if len(max_p) != len(dimensions):
                raise AssertionError(
                    "max_p must be a list of length equal to the number of dimensions"
                )
        else:
            max_p = [max_p] * len(dimensions)

        max_p_ = [0, 0, 0, 0]
        max_b_ = [0, 0, 0, 0]

        for i, dim in enumerate(dimensions):
            max_p_[dim] = max_p[i]
            max_b_[dim] = max_b[i]

        self.max_b = max_b_
        self.max_p = max_p_

        self.max_fit_tolerance = max_fit_tolerence
        self.periodic = periodic
        self.no_supercell = no_supercell
        self.alpha_weight = alpha_weight

        super().__init__(primitive=primitive)

    def _get_feature_labels(self) -> List[str]:
        labels = []
        _elements = list(self.atom_types)
        if self.compute_for_all_elements:
            _elements.append("all")
        for element in _elements:
            for dim in self.dimensions:
                for pixel_a in range(self.image_size[0]):
                    for pixel_b in range(self.image_size[1]):
                        labels.append(f"phimage_{element}_{dim}_{pixel_a}_{pixel_b}")

        return labels

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def _featurize(
        self, structure: Union[Structure, IStructure, Molecule, IMolecule]
    ) -> np.ndarray:
        results = get_persistent_images_for_structure(
            structure,
            elements=self.atom_types,
            compute_for_all_elements=self.compute_for_all_elements,
            min_size=self.min_size,
            pixels=self.image_size,
            spread=self.spread,
            weighting=self.weight,
            max_b=self.max_b,
            max_p=self.max_p,
            periodic=self.periodic,
            no_supercell=self.no_supercell,
            alpha_weighting=self.alpha_weight,
        )
        features = []
        elements = list(self.atom_types)
        if self.compute_for_all_elements:
            elements.append("all")
        for element in elements:
            for dim in self.dimensions:

                features.append(np.array(results["image"][element][dim]).flatten())
        return np.concatenate(features)

    def _fit(self, structures: List[Union[Structure, IStructure, Molecule, IMolecule]]) -> None:
        """Use structures to estimate the settings for the featurizer.

        Find the limits (maximum/minimum birth/death and persistence)
        for all the structures in the dataset and store them in the object.

        Args:
            structures (List[Union[Structure, IStructure, Molecule, IMolecule]]): List of structures
                to find the limits for.
        """
        limits = defaultdict(list)

        for structure in structures:
            lim = get_persistence_image_limits_for_structure(
                structure,
                self.atom_types,
                self.compute_for_all_elements,
                self.min_size,
                periodic=self.periodic,
                no_supercell=self.no_supercell,
                alpha_weighting=self.alpha_weight,
            )
            for k, v in lim.items():
                limits[k].extend(v)

        # birth min, max persistence min, max
        maxp = []
        maxb = []

        for _, v in limits.items():
            v = np.array(v)
            mb = np.max(v[:, 1])
            mp = np.max(v[:, 3])
            maxb.append(mb + self.max_fit_tolerance * mb)
            maxp.append(mp + self.max_fit_tolerance * mp)

        self.max_b = maxb
        self.max_p = maxp

    def citations(self) -> List[str]:
        return [
            "@article{doi:10.1021/acs.jpcc.0c01167,"
            "author = {Krishnapriyan, Aditi S. and Haranczyk, Maciej and Morozov, Dmitriy},"
            "title = {Topological Descriptors Help Predict Guest Adsorption in Nanoporous Materials},"
            "journal = {The Journal of Physical Chemistry C},"
            "volume = {124},"
            "number = {17},"
            "pages = {9360-9368},"
            "year = {2020},"
            "doi = {10.1021/acs.jpcc.0c01167},"
            "}",
            "@article{krishnapriyan_machine_2021,"
            "title={Machine learning with persistent homology and chemical word embeddings "
            "improves prediction accuracy and interpretability in metal-organic frameworks},"
            r"author={Krishnapriyan, Aditi S and Montoya, Joseph and Haranczyk, Maciej and "
            r"Hummelsh{\o}j, Jens and Morozov, Dmitriy},"
            "journal = {Scientific Reports},"
            "volume = {11},"
            "numer = {1},"
            "issn = {2045-2322},"
            "pages = {8888},"
            "year={2021},"
            "doi = {10.1038/s41598-021-88027-8}"
            "}",
            "@article{adams2017persistence,"
            "title={Persistence images: A stable vector representation of persistent homology},"
            "author={Adams, Henry and Emerson, Tegan and Kirby, Michael and Neville, Rachel "
            "and Peterson, Chris and Shipman, Patrick and Chepushtanova, Sofya and Hanson, "
            "Eric and Motta, Francis and Ziegelmeier, Lori},"
            "journal={Journal of Machine Learning Research},"
            "volume={18},"
            "year={2017}"
            "}",
        ]

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka", "Aditi Krishnapriyan"]



# -*- coding: utf-8 -*-
"""Computing pore properties using Zeo++."""
import os
import re
import subprocess
from io import StringIO
from tempfile import TemporaryDirectory
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from loguru import logger
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer

from ..utils import is_tool
from ..utils.tempdir import TEMPDIR

ZEOPP_BASE_COMMAND = ["network"]
HA_COMMAND = ["-ha"]
NO_ZEOPP_WARNING = "Did not find the zeo++ network binary in the path. \
            Can not run pore analysis."


PROBE_RADII = {
    "CO2": 1.525,
    "N2": 1.655,
    "H2": 1.48,
    "CH4": 1.865,
    "O2": 1.51,
    "Xe": 1.985,
    "Kr": 1.83,
    "H2O": 1.58,
    "H2S": 1.74,
    "C6H6": 2.925,
}

__all__ = [
    "AccessibleVolume",
    "PoreDiameters",
    "PoreSizeDistribution",
    "RayTracingHistogram",
    "SurfaceArea",
]


def run_zeopp(structure: Structure, command: str, parser: Callable, ha: bool = True) -> dict:
    """Run zeopp with network -ha to find the pore diameters.

    Args:
        structure (Structure): pymatgen Structure object
        command (str): command for zeopp
        parser (Callable): function to parse the output of zeopp
        ha (bool): whether to use the 'high accuracy' :code:`-ha` flag

    Returns:
        dict: pore analysis results
    """
    if not is_tool("network"):
        logger.error(NO_ZEOPP_WARNING)
    with TemporaryDirectory(dir=TEMPDIR) as tempdir:
        structure_path = os.path.join(tempdir, "structure.cif")
        result_path = os.path.join(tempdir, "result.res")
        structure.to(filename=structure_path, fmt="cif")
        if ha:
            cmd = (
                ZEOPP_BASE_COMMAND + HA_COMMAND + command + [str(result_path), str(structure_path)]
            )
        else:
            cmd = ZEOPP_BASE_COMMAND + command + [str(result_path), str(structure_path)]
        _ = subprocess.run(  # nosec
            cmd,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
            cwd=tempdir,
        )

        with open(result_path, "r") as handle:
            results = handle.read()

        zeopp_results = parser(results)

        return zeopp_results


def _parse_res_zeopp(filecontent: str) -> Tuple[List[float], List[str]]:
    """Parse the results line of a network -res call to zeopp

    Args:
        filecontent (str): results file

    Returns:
        dict: largest included sphere, largest free sphere,
            largest included sphere along free sphere path
    """
    first_line = filecontent.split("\n")[0]
    parts = first_line.split()

    results = {
        "lis": float(parts[1]),  # largest included sphere
        "lifs": float(parts[2]),  # largest free sphere
        "lifsp": float(parts[3]),  # largest included sphere along free sphere path
    }

    return results


def _parse_sa_zeopp(filecontent):
    regex_unitcell = re.compile(r"Unitcell_volume: ((\d+\.\d+)|\d+)|$")
    regex_density = re.compile(r"Density: ((\d+\.\d+)|\d+)|$")
    asa_a2 = re.compile(r"ASA_A\^2: ((\d+\.\d+)|\d+)|$")
    asa_m2cm3 = re.compile(r"ASA_m\^2/cm\^3: ((\d+\.\d+)|\d+)|$")
    asa_m2g = re.compile(r"ASA_m\^2/g: ((\d+\.\d+)|\d+)|$")
    nasa_a2 = re.compile(r"NASA_A\^2: ((\d+\.\d+)|\d+)|$")
    nasa_m2cm3 = re.compile(r"NASA_m\^2/cm\^3: ((\d+\.\d+)|\d+)|$")
    nasa_m2g = re.compile(r"NASA_m\^2/g: ((\d+\.\d+)|\d+)|$")

    d = {
        "unitcell_volume": float(re.findall(regex_unitcell, filecontent)[0][0]),
        "density": float(re.findall(regex_density, filecontent)[0][0]),
        "asa_a2": float(re.findall(asa_a2, filecontent)[0][0]),
        "asa_m2cm3": float(re.findall(asa_m2cm3, filecontent)[0][0]),
        "asa_m2g": float(re.findall(asa_m2g, filecontent)[0][0]),
        "nasa_a2": float(re.findall(nasa_a2, filecontent)[0][0]),
        "nasa_m2cm3": float(re.findall(nasa_m2cm3, filecontent)[0][0]),
        "nasa_m2g": float(re.findall(nasa_m2g, filecontent)[0][0]),
    }

    return d


def _parse_volpo_zeopp(filecontent):
    regex_unitcell = re.compile(r"Unitcell_volume: ((\d+\.\d+)|\d+)|$")
    regex_density = re.compile(r"Density: ((\d+\.\d+)|\d+)|$")
    av_a3 = re.compile(r"AV_A\^3: ((\d+\.\d+)|\d+)|$")
    av_volume_fraction = re.compile(r"AV_Volume_fraction: ((\d+\.\d+)|\d+)|$")
    av_cm3g = re.compile(r"AV_cm\^3/g: ((\d+\.\d+)|\d+)|$")
    nav_a3 = re.compile(r"NAV_A\^3: ((\d+\.\d+)|\d+)|$")
    nav_volume_fraction = re.compile(r"NAV_Volume_fraction: ((\d+\.\d+)|\d+)|$")
    nav_cm3g = re.compile(r"NAV_cm\^3/g: ((\d+\.\d+)|\d+)|$")

    d = {
        "unitcell_volume": float(re.findall(regex_unitcell, filecontent)[0][0]),
        "density": float(re.findall(regex_density, filecontent)[0][0]),
        "av_a3": float(re.findall(av_a3, filecontent)[0][0]),
        "av_volume_fraction": float(re.findall(av_volume_fraction, filecontent)[0][0]),
        "av_cm3g": float(re.findall(av_cm3g, filecontent)[0][0]),
        "nav_a3": float(re.findall(nav_a3, filecontent)[0][0]),
        "nav_volume_fraction": float(re.findall(nav_volume_fraction, filecontent)[0][0]),
        "nav_cm3g": float(re.findall(nav_cm3g, filecontent)[0][0]),
    }

    return d


def _parse_ray_hist_zeopp(filecontent):
    return [float(i) for i in filecontent.split("\n")[1:-1]]


def _parse_psd_zeopp(filecontent):
    return pd.read_csv(
        StringIO(filecontent),
        skiprows=11,
        names=["bin", "count", "cumulative", "derivative"],
        sep=r"\s+",
    )


class PoreDiameters(MOFBaseFeaturizer):
    """Calculate the pore diameters of a framework."""

    def __init__(
        self,
        ha: bool = True,
        primitive: bool = True,
    ):
        """Initialize the featurizer.

        Args:
            ha (bool): if True, run zeo++ with the "high accuracy"
                :code:`-ha` flag.
                It has been reported that this can lead to issues
                for some structures.
                Default is True.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        self.labels = ["lis", "lifs", "lifsp"]
        self.ha = ha
        super().__init__(primitive=primitive)

    def _featurize(self, s):
        result = run_zeopp(s, ["-res"], _parse_res_zeopp, self.ha)
        return np.array(list(result.values()))

    def feature_labels(self):
        return self.labels

    def citations(self):
        return [
            "@article{Willems2012,"
            "doi = {10.1016/j.micromeso.2011.08.020},"
            "url = {https://doi.org/10.1016/j.micromeso.2011.08.020},"
            "year = {2012},"
            "month = feb,"
            "publisher = {Elsevier {BV}},"
            "volume = {149},"
            "number = {1},"
            "pages = {134--141},"
            "author = {Thomas F. Willems and Chris H. Rycroft and Michaeel Kazi "
            "and Juan C. Meza and Maciej Haranczyk},"
            "title = {Algorithms and tools for high-throughput geometry-based "
            "analysis of crystalline porous materials},"
            "journal = {Microporous and Mesoporous Materials}"
            "}"
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka", "Maciej Haranczyk and Zeo++ authors"]


class SurfaceArea(MOFBaseFeaturizer):
    def __init__(
        self,
        probe_radius: Union[str, float] = 0.1,
        num_samples: int = 100,
        channel_radius: Union[str, float, None] = None,
        ha: bool = True,
        primitive: bool = True,
    ):
        """Initialize the SurfaceArea featurizer.

        Args:
            probe_radius (Union[str, float]): Radius of the probe.
                Defaults to 0.1.
            num_samples (int): Number of samples.
                Defaults to 100.
            channel_radius (Union[str, float, None]): Channel radius.
                Should equal to `probe_radius`. Defaults to None.
            ha (bool): if True, run zeo++ with the "high accuracy"
                :code:`-ha` flag.
                It has been reported that this can lead to issues
                for some structures.
                Default is True.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        if channel_radius is not None and probe_radius != channel_radius:
            logger.warning(
                "Probe radius and channel radius are different. This is a highly unusual setting."
            )
        if isinstance(probe_radius, str):
            try:
                probe_radius = PROBE_RADII[probe_radius]
            except KeyError:
                logger.error(f"Probe radius {probe_radius} not found in PROBE_RADII")

        if channel_radius is None:
            channel_radius = probe_radius

        self.ha = ha
        self.probe_radius = probe_radius
        self.num_samples = num_samples
        self.channel_radius = channel_radius

        labels = [
            "uc_volume",
            "density",
            "asa_a2",
            "asa_m2cm3",
            "asa_m2g",
            "nasa_a2",
            "nasa_m2cm3",
            "nasa_m2g",
        ]

        self.labels = [f"{label}_{self.probe_radius}" for label in labels]
        super().__init__(primitive=primitive)

    def _featurize(self, s: Union[Structure, IStructure]) -> np.ndarray:
        command = [
            "-sa",
            f"{self.channel_radius}",
            f"{self.probe_radius}",
            f"{self.num_samples}",
        ]
        results = run_zeopp(s, command, _parse_sa_zeopp, self.ha)
        return np.array(list(results.values()))

    def feature_labels(self) -> List[str]:
        return self.labels

    def citations(self) -> List[str]:
        return [
            "@article{Willems2012,"
            "doi = {10.1016/j.micromeso.2011.08.020},"
            "url = {https://doi.org/10.1016/j.micromeso.2011.08.020},"
            "year = {2012},"
            "month = feb,"
            "publisher = {Elsevier {BV}},"
            "volume = {149},"
            "number = {1},"
            "pages = {134--141},"
            "author = {Thomas F. Willems and Chris H. Rycroft and Michaeel Kazi "
            "and Juan C. Meza and Maciej Haranczyk},"
            "title = {Algorithms and tools for high-throughput geometry-based "
            "analysis of crystalline porous materials},"
            "journal = {Microporous and Mesoporous Materials}"
            "}"
        ]

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka", "Maciej Haranczyk and Zeo++ authors"]


class AccessibleVolume(MOFBaseFeaturizer):
    def __init__(
        self,
        probe_radius: Union[str, float] = 0.1,
        num_samples: int = 100,
        channel_radius: Union[str, float, None] = None,
        ha: bool = True,
        primitive: bool = True,
    ):
        """Initialize the AccessibleVolume featurizer.

        Args:
            probe_radius (Union[str, float]): Radius of the probe.
                Defaults to 0.1.
            num_samples (int): Number of samples.
                Defaults to 100.
            channel_radius (Union[str, float, None]): Channel radius.
                Should equal to `probe_radius`. Defaults to None.
            ha (bool): if True, run zeo++ with the "high accuracy"
                :code:`-ha` flag.
                It has been reported that this can lead to issues
                for some structures.
                Default is True.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        if channel_radius is not None and probe_radius != channel_radius:
            logger.warning(
                "Probe radius and channel radius are different. This is a highly unusual setting."
            )
        if isinstance(probe_radius, str):
            try:
                probe_radius = PROBE_RADII[probe_radius]
            except KeyError:
                logger.error(f"Probe radius {probe_radius} not found in PROBE_RADII")

        if channel_radius is None:
            channel_radius = probe_radius

        self.ha = ha
        self.probe_radius = probe_radius
        self.num_samples = num_samples
        self.channel_radius = channel_radius
        labels = [
            "uc_volume",
            "density",
            "av_a2",
            "av_volume_fraction",
            "av_cm3g",
            "nav_a3",
            "nav_volume_fraction",
            "nav_cm3g",
        ]
        self.labels = [f"{label}_{self.probe_radius}" for label in labels]
        super().__init__(primitive=primitive)

    def _featurize(self, s: Union[Structure, IStructure]) -> np.ndarray:
        command = ["-vol", f"{self.channel_radius}", f"{self.probe_radius}", f"{self.num_samples}"]
        results = run_zeopp(s, command, _parse_volpo_zeopp, self.ha)
        return np.array(list(results.values()))

    def feature_labels(self) -> List[str]:
        return self.labels

    def citations(self) -> List[str]:
        return [
            "@article{Willems2012,"
            "doi = {10.1016/j.micromeso.2011.08.020},"
            "url = {https://doi.org/10.1016/j.micromeso.2011.08.020},"
            "year = {2012},"
            "month = feb,"
            "publisher = {Elsevier {BV}},"
            "volume = {149},"
            "number = {1},"
            "pages = {134--141},"
            "author = {Thomas F. Willems and Chris H. Rycroft and Michaeel Kazi "
            "and Juan C. Meza and Maciej Haranczyk},"
            "title = {Algorithms and tools for high-throughput geometry-based "
            "analysis of crystalline porous materials},"
            "journal = {Microporous and Mesoporous Materials}"
            "}",
            "@article{Ongari2017,"
            "doi = {10.1021/acs.langmuir.7b01682},"
            "url = {https://doi.org/10.1021/acs.langmuir.7b01682},"
            "year = {2017},"
            "month = jul,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {33},"
            "number = {51},"
            "pages = {14529--14538},"
            "author = {Daniele Ongari and Peter G. Boyd and Senja Barthel "
            "and Matthew Witman and Maciej Haranczyk and Berend Smit},"
            "title = {Accurate Characterization of the Pore Volume in "
            "Microporous Crystalline Materials},"
            "journal = {Langmuir}"
            "}",
        ]

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka", "Maciej Haranczyk and Zeo++ authors"]


class RayTracingHistogram(MOFBaseFeaturizer):
    """Describe pore structures using histograms of ray lengths.

    The algorithm (implemented in `zeo++ <http://www.zeoplusplus.org/>`_)
    shoots random rays through the accesible volume of the cell until the ray
    hits atoms, and it records their lenghts to provide the corresponding
    histogram.

    Such ray histograms are supposed to encode the shape, topology, distribution
    and size of voids.

    Currently, the histogram is hard-coded to be of length 1000 (in zeo++
    itself).
    """

    def __init__(
        self,
        probe_radius: Union[str, float] = 0.0,
        num_samples: int = 50000,
        channel_radius: Optional[Union[str, float]] = None,
        ha: bool = True,
        primitive: bool = True,
    ) -> None:
        """Initialize the RayTracingHistogram featurizer.

        Args:
            probe_radius (Union[str, float]): Used to estimate the accessible volume.
                Only the accessible volume is then considered for the histogram.
                Defaults to 0.0.
            num_samples (int): Number of rays that are placed through sample.
                Original publication used  1,000,000 sample points for IZA zeolites
                and 100,000 sample points for hypothetical zeolites.
                Larger numbers increase the runtime Defaults to 50000.
            channel_radius (Union[str, float, None]):  Radius of a probe
                used to determine accessibility of the void space.
                Should typically equal the radius of the `probe_radius`.
                If set to `None`, we will use the `probe_radius`. Defaults to None.
            ha (bool): if True, run zeo++ with the "high accuracy"
                :code:`-ha` flag.
                It has been reported that this can lead to issues
                for some structures.
                Default is True.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        if channel_radius is not None and probe_radius != channel_radius:
            logger.warning(
                "Probe radius and channel radius are different. This is a highly unusual setting."
            )
        if isinstance(probe_radius, str):
            try:
                probe_radius = PROBE_RADII[probe_radius]
            except KeyError:
                logger.error(f"Probe radius {probe_radius} not found in PROBE_RADII")

        if channel_radius is None:
            channel_radius = probe_radius

        self.probe_radius = probe_radius
        self.num_samples = num_samples
        self.channel_radius = channel_radius
        self.ha = ha
        super().__init__(primitive=primitive)

    def feature_labels(self) -> List[str]:
        return [f"ray_hist_{self.probe_radius}_{i}" for i in range(1000)]

    def _featurize(self, s: Union[Structure, IStructure]) -> np.ndarray:
        command = [
            "-ray_atom",
            f"{self.channel_radius}",
            f"{self.probe_radius}",
            f"{self.num_samples}",
        ]
        results = run_zeopp(s, command, _parse_ray_hist_zeopp, self.ha)
        return np.array(results)

    def citations(self) -> List[str]:
        return [
            "@article{Jones2013,"
            "doi = {10.1016/j.micromeso.2013.07.033},"
            "url = {https://doi.org/10.1016/j.micromeso.2013.07.033},"
            "year = {2013},"
            "month = nov,"
            "publisher = {Elsevier {BV}},"
            "volume = {181},"
            "pages = {208--216},"
            "author = {Andrew J. Jones and Christopher Ostrouchov and Maciej Haranczyk "
            "and Enrique Iglesia},"
            "title = {From rays to structures: Representation and selection of "
            "void structures in zeolites using stochastic methods},"
            "journal = {Microporous and Mesoporous Materials}"
            "}",
            "@article{Willems2012,"
            "doi = {10.1016/j.micromeso.2011.08.020},"
            "url = {https://doi.org/10.1016/j.micromeso.2011.08.020},"
            "year = {2012},"
            "month = feb,"
            "publisher = {Elsevier {BV}},"
            "volume = {149},"
            "number = {1},"
            "pages = {134--141},"
            "author = {Thomas F. Willems and Chris H. Rycroft and Michaeel Kazi "
            "and Juan C. Meza and Maciej Haranczyk},"
            "title = {Algorithms and tools for high-throughput geometry-based "
            "analysis of crystalline porous materials},"
            "journal = {Microporous and Mesoporous Materials}"
            "}",
        ]

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka", "Maciej Haranczyk and Zeo++ authors"]


class PoreSizeDistribution(MOFBaseFeaturizer):
    """Describe structures using histograms of pore sizes.

    The pore size distribution describes how much of the void space
    corresponds to certain pore sizes.

    Pinheiro et al. (2013) concluded that they are "sensitive to small changes
    in pore diameter" but do "not reflect subtle changes in features such as the
    surface texture of a pore".

    We use the implementation in `zeo++ <http://www.zeoplusplus.org/>`_ to
    calculate the pore size distribution.

    The pore size distribution has been used by the group of Gómez-Gualdrón  as
    pore size standard deviation (PSSD) in, for example,
    `10.1021/acs.jctc.9b00940
    <https://pubs.acs.org/doi/10.1021/acs.jctc.9b00940>`_ and `10.1063/5.0048736
    <https://aip.scitation.org/doi/10.1063/5.0048736>`_.

    Currently, the histogram is hard-coded to be of length 1000 between 0 and
    100 Angstrom (in zeo++ itself).
    """

    def __init__(
        self,
        probe_radius: Union[str, float] = 0.0,
        num_samples: int = 5000,
        channel_radius: Optional[Union[str, float]] = None,
        hist_type: str = "derivative",
        ha: bool = False,
        primitive: bool = True,
    ) -> None:
        """Initialize the PoreSizeDistribution featurizer.

        Args:
            probe_radius (Union[str, float]): Used to estimate the accessible volume.
                Only the accessible volume is then considered for the histogram.
                Defaults to 0.0.
            num_samples (int): Number of rays that are placed through sample.
                Original publication used  1,000,000 sample points for IZA zeolites and 100,000 sample points
                for hypothetical zeolites. Larger numbers increase the runtime. Defaults to 50000.
            channel_radius (Union[str, float, None]): Radius of a probe used to determine
                accessibility of the void space. Should typically equal the radius of the `probe_radius`.
                If set to `None`, we will use the `probe_radius`. Defaults to None.
            hist_type (str): Type of the histogram.
                Available options `count`, `cumulative`, `derivative`.
                (The derivative distribution describes the change in the cumulative distribution
                with respect to pore size). Defaults to "derivative".
            ha (bool): if True, run zeo++ with the "high accuracy"
                :code:`-ha` flag.
                It has been reported that this can lead to issues
                for some structures.
                Default is True.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.

        Raises:
            ValueError: If type not one of 'count', 'cumulative', 'derivative'.
        """
        if channel_radius is not None and probe_radius != channel_radius:
            logger.warning(
                "Probe radius and channel radius are different. This is a highly unusual setting."
            )
        if isinstance(probe_radius, str):
            try:
                probe_radius = PROBE_RADII[probe_radius]
            except KeyError:
                logger.error(f"Probe radius {probe_radius} not found in PROBE_RADII")

        if channel_radius is None:
            channel_radius = probe_radius

        self.type = hist_type.lower()
        if self.type not in [
            "count",
            "cumulative",
            "derivative",
        ]:
            raise ValueError(
                "Invalid histogram type, must be one of `count`, `cumulative`, `derivative`"
            )

        self.probe_radius = probe_radius
        self.num_samples = num_samples
        self.channel_radius = channel_radius
        self.ha = ha
        super().__init__(primitive=primitive)

    def feature_labels(self) -> List[str]:
        return [f"psd_hist_{self.probe_radius}_{i}" for i in range(1000)]

    def _featurize(self, s: Union[Structure, IStructure]) -> np.ndarray:
        command = [
            "-psd",
            f"{self.channel_radius}",
            f"{self.probe_radius}",
            f"{self.num_samples}",
        ]
        results = run_zeopp(s, command, _parse_psd_zeopp, self.ha)
        return results[self.type].values

    def citations(self) -> List[str]:
        return [
            "@article{Pinheiro2013,"
            "doi = {10.1016/j.jmgm.2013.05.007},"
            "url = {https://doi.org/10.1016/j.jmgm.2013.05.007},"
            "year = {2013},"
            "month = jul,"
            "publisher = {Elsevier {BV}},"
            "volume = {44},"
            "pages = {208--219},"
            "author = {Marielle Pinheiro and Richard L. Martin and "
            "Chris H. Rycroft and Andrew Jones and Enrique Iglesia and Maciej Haranczyk},"
            "title = {Characterization and comparison of pore landscapes "
            "in crystalline porous materials},"
            "journal = {Journal of Molecular Graphics and Modelling}"
            "}",
            "@article{Willems2012,"
            "doi = {10.1016/j.micromeso.2011.08.020},"
            "url = {https://doi.org/10.1016/j.micromeso.2011.08.020},"
            "year = {2012},"
            "month = feb,"
            "publisher = {Elsevier {BV}},"
            "volume = {149},"
            "number = {1},"
            "pages = {134--141},"
            "author = {Thomas F. Willems and Chris H. Rycroft and Michaeel Kazi "
            "and Juan C. Meza and Maciej Haranczyk},"
            "title = {Algorithms and tools for high-throughput geometry-based "
            "analysis of crystalline porous materials},"
            "journal = {Microporous and Mesoporous Materials}"
            "}",
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka", "Maciej Haranczyk and Zeo++ authors"]



# -*- coding: utf-8 -*-
"""Atomic-property weighted autocorrelation function.

See alternative implementation https://github.com/tomdburns/AP-RDF (likely
faster as it also has a lower-level implementation)
"""
from functools import cached_property
from typing import List, Tuple, Union

import numpy as np
from element_coder import encode
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer

from ..utils.aggregators import AGGREGATORS
from ..utils.extend import operates_on_istructure, operates_on_structure

__all__ = ["APRDF"]


@operates_on_structure
@operates_on_istructure
class APRDF(MOFBaseFeaturizer):
    r"""Generalization of descriptor described by `Fernandez et al. <https://pubs.acs.org/doi/10.1021/jp404287t>`_.

    In the article they describe the product of atomic properties as weighting
    of a "conventional" radiual distribution function "RDF".

    .. math::
        \operatorname{RDF}^{p}(R)=f \sum_{i, j}^{\text {all atom puirs }} P_{i} P_{j} \mathrm{e}^{-B\left(r_{ij}-R\right)^{2}} # noqa: E501

    Here, we allow for a wider choice of option for aggregation of properties
    :math:`P_{i}` and :math:`P_{j}` (not only the product).

    Examples:
        >>> from mofdscribe.chemistry.aprdf import APRDF
        >>> from pymatgen.core.structure import Structure
        >>> s = Structure.from_file("tests/test_files/LiTiO3.cif")
        >>> aprdf = APRDF()
        >>> features = aprdf.featurize(s)
    """

    def __init__(
        self,
        cutoff: float = 20.0,
        lower_lim: float = 2.0,
        bin_size: float = 0.25,
        b_smear: Union[float, None] = 10,
        properties: Tuple[str, int] = ("X", "electron_affinity"),
        aggregations: Tuple[str] = ("avg", "product", "diff"),
        normalize: bool = False,
        primitive: bool = True,
    ):
        """Set up an atomic property (AP) weighted radial distribution function.

        Args:
            cutoff (float): Consider neighbors up to this value (in
                Angstrom). Defaults to 20.0.
            lower_lim (float): Lowest distance (in Angstrom) to consider.
                Defaults to 2.0.
            bin_size (float): Bin size for binning.
                Defaults to 0.25.
            b_smear (Union[float, None]): Band width for Gaussian smearing.
                If None, the unsmeared histogram is used. Defaults to 10.
            properties (Tuple[str, int]): Properties used for calculation of the AP-RDF.
                All properties of `pymatgen.core.Species` are available in
                addition to the integer `1` that will set P_i=P_j=1. Defaults to
                ("X", "electron_affinity").
            aggregations (Tuple[str]): Methods used to combine the
                properties.
                See `mofdscribe.featurizers.utils.aggregators.AGGREGATORS` for available
                options. Defaults to ("avg", "product", "diff").
            normalize (bool): If True, the histogram is normalized by dividing
                by the number of atoms. Defaults to False.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        self.lower_lim = lower_lim
        self.cutoff = cutoff
        self.bin_size = bin_size
        self.properties = properties
        self.normalize = normalize

        self.b_smear = b_smear
        self.aggregations = aggregations
        super().__init__(primitive=primitive)

    def precheck(self):
        pass

    @cached_property
    def _bins(self):
        num_bins = int((self.cutoff - self.lower_lim) // self.bin_size)
        bins = np.linspace(self.lower_lim, self.cutoff, num_bins)
        return bins

    def _get_feature_labels(self):
        aprdfs = np.empty(
            (len(self.properties), len(self.aggregations), len(self._bins)), dtype=object
        )
        for pi, prop in enumerate(self.properties):
            for ai, aggregation in enumerate(self.aggregations):
                for bin_index, _ in enumerate(self._bins):
                    aprdfs[pi][ai][bin_index] = f"aprdf_{prop}_{aggregation}_{bin_index}"

        return list(aprdfs.flatten())

    def _featurize(self, s: Union[Structure, IStructure]) -> np.array:
        bins = self._bins
        aprdfs = np.zeros((len(self.properties), len(self.aggregations), len(bins)))

        # todo: use numba to speed up
        for i, item in enumerate(s):
            for j in range(i + 1, len(s)):
                dist = s.get_distance(i, j)
                if dist < self.cutoff and dist > self.lower_lim:
                    bin_idx = int((dist - self.lower_lim) // self.bin_size)
                    for pi, prop in enumerate(self.properties):
                        for ai, agg in enumerate(self.aggregations):
                            p0 = encode(item.specie, prop)
                            p1 = encode(s[j].specie, prop)

                            agg_func = AGGREGATORS[agg]
                            p = agg_func([p0, p1])
                            aprdfs[pi][ai][bin_idx] += p * np.exp(
                                -self.b_smear * (dist - bins[bin_idx]) ** 2
                            )

        if self.normalize:
            aprdfs /= len(s)

        return aprdfs.flatten()

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def citations(self) -> List[str]:
        return [
            "@article{Fernandez2013,"
            "doi = {10.1021/jp404287t},"
            "url = {https://doi.org/10.1021/jp404287t},"
            "year = {2013},"
            "month = jul,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {117},"
            "number = {27},"
            "pages = {14095--14105},"
            "author = {Michael Fernandez and Nicholas R. Trefiak and Tom K. Woo},"
            "title = {Atomic Property Weighted Radial Distribution Functions "
            "Descriptors of Metal{\textendash}Organic Frameworks for the Prediction "
            "of Gas Uptake Capacity},"
            "journal = {The Journal of Physical Chemistry C}"
            "}"
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka"]



# -*- coding: utf-8 -*-
"""Measure the RMSD between a building block and topological prototypes."""
import json
import os
from collections import OrderedDict
from typing import List, Tuple, Union

import numpy as np
from loguru import logger
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from ..utils.aggregators import ARRAY_AGGREGATORS

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(THIS_DIR, "prototype_env.json"), "r") as handle:
    STRUCTURE_ENVS = json.loads(handle.read())

ALL_AVAILABLE_TOPOS = tuple(STRUCTURE_ENVS.keys())

__all__ = ("BUMatch",)


def match_bb(
    bb: Union[Structure, IStructure, Molecule, IMolecule],
    prototype: str,
    aggregations: Tuple[str],
    allow_rescale: bool = True,
    match: str = "auto",
    skip_non_fitting_if_possible: bool = True,
    mismatch_fill_value: float = 10_000,
) -> float:
    """
    Compute the RMSD between a building block and a prototype.

    Args:
        bb (Union[Structure, IStructure, Molecule, IMolecule]): The building block to compare.
        prototype (str): The prototype to compare against.
        aggregations (Tuple[str]): The aggregations to use.
        allow_rescale (bool): Whether to scale the RMSD by the number of atoms.
            Defaults to True.
        match (str): The type of matching to use. Defaults to 'auto'.
        skip_non_fitting_if_possible (bool): Whether to skip RMSDs of
            building blocks that do not match due to mismatching coordination numbers.
        mismatch_fill_value (float): The value to fill in for mismatching
            coordination numbers. Defaults to 10_000.

    Returns:
        The RMSD between the two structures.
    """
    from superpose3d import Superpose3D

    if match != "auto":
        is_node = 1 if match == "node" else -1
    else:
        cn = len(bb)
        is_node = -1 if cn == 2 else 1
    logger.debug(f"Matching {bb} to {prototype} with {aggregations}. Is node: {is_node}")
    coords_this = bb.cart_coords
    keys_to_match = [k for k in STRUCTURE_ENVS[prototype].keys() if int(k) * (is_node) >= 0]
    logger.debug(f"Matching {keys_to_match}")
    rmsds_fitting = []
    rmsds_non_fitting = []
    for key in keys_to_match:
        reference_coordinates = STRUCTURE_ENVS[prototype][key]
        reference_coordinates = np.array(reference_coordinates)

        if len(reference_coordinates) == len(coords_this):
            rmsd, _, _, _ = Superpose3D(
                reference_coordinates, coords_this, allow_rescale=allow_rescale
            )
            rmsds_fitting.append(rmsd)
        else:
            rmsds_non_fitting.append(mismatch_fill_value)
    logger.debug(f"Fitting RMSDs: {rmsds_fitting}")
    logger.debug(f"Non-fitting RMSDs: {rmsds_non_fitting}")
    rmsds = None
    if (len(rmsds_fitting) > 0) & skip_non_fitting_if_possible:
        rmsds = rmsds_fitting
    else:
        rmsds = rmsds_non_fitting + rmsds_fitting
    aggregation_results = OrderedDict()

    for aggregation in aggregations:
        aggregation_results[f"{prototype}_{aggregation}"] = ARRAY_AGGREGATORS[aggregation](rmsds)

    return aggregation_results


class BUMatch(BaseFeaturizer):
    """MOFs are assembled from building blocks on a net.

    The "ideal" vertex "structures" of the net can fit better or
    worse with the "shape" of the actual building blocks.
    This featurizer attempts to quantify this mismatch.

    .. note::

        The edge match values will all be quite close to zero
        and hence not that meaningful (two points always form a line,
        there is not much room for mismatch unless the length of the line
        [which is ignored by default with `allow_rescale=True`]...).
        In pratice, you should consider treating them seperate
        from the vertex match values.

    Examples:
        >>> from mofdscribe.bu import BUMatch
        >>> from pymatgen.core import Structure
        >>> s = Structure.from_file("tests/test_files/bu_test_1.cif")
        >>> bu_match = BUMatch(topos=["tbo", "pcu"], aggregations=["mean", "min"])
        >>> bu_match.featurize(s)
    """

    def __init__(
        self,
        allow_rescale: bool = True,
        mismatch_fill_value: float = 1_000,
        return_only_best: bool = True,
        aggregations: Tuple[str] = ("max", "min", "mean", "std"),
        topos: Tuple[str] = ALL_AVAILABLE_TOPOS,
        match: str = "auto",
        skip_non_fitting_if_possible: bool = True,
    ) -> None:
        """Create a new BUMatch featurizer.

        Args:
            allow_rescale (bool): If True, allow to multiple coordinates of structure
                with scalar to better match the reference structure.
                Defaults to True.
            mismatch_fill_value (float): Value use to fill entries for which the RMSD
                computation cannot be perform due to a mismatch in coordination numbers.
                Defaults to 1_000.
            return_only_best (bool): If True, do not compute statistics but only return the minimum
                RMSD. Defaults to True.
            aggregations (Tuple[str]): Functions to use to aggregate RMSD of
                the different possible positions. Defaults to ("max", "min", "mean", "std").
            topos (Tuple[str]): RCSR codes to consider for matching.
                Defaults to ALL_AVAILABLE_TOPOS.
            match (str): BB to consider for matching. Must be one of "auto", "edge" or "node".
                If the mode is "auto" it assumes that the number of sites in the input building
                block is equal to the number of connection vertices. Hence, if there are more
                than two sites, it will match the node. If there are only two sites, it will match
                the edge.
                Defaults to "auto".
            skip_non_fitting_if_possible (bool): If True, do not compute RMSD for
                non-compatible BBs. Defaults to True.
        """
        self.allow_rescale = allow_rescale
        self.mismatch_fill_value = mismatch_fill_value
        self.topos = topos
        self.return_only_best = return_only_best
        if not return_only_best and aggregations is None:
            logger.error("If return_only_best is False, aggregations must be set.")
        self.aggregations = aggregations
        if self.return_only_best:
            self.aggregations = ("min",)

        self.match = match
        self.skip_non_fitting_if_possible = skip_non_fitting_if_possible

    def _get_feature_labels(self):
        labels = []
        for topo in self.topos:
            if self.return_only_best:
                labels.append(f"bumatch_{self.allow_rescale}_{topo}")
            else:
                for aggregation in self.aggregations:
                    labels.append(f"bumatch_{self.allow_rescale}_{topo}_{aggregation}")

        return labels

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def featurize(self, s: Union[Structure, IStructure, Molecule, IMolecule]) -> np.ndarray:
        """Structure is here spanned by the connecting points of a BU."""
        features = []
        for topo in self.topos:
            feats = match_bb(
                s,
                topo,
                self.aggregations,
                self.allow_rescale,
                self.match,
                self.skip_non_fitting_if_possible,
                self.mismatch_fill_value,
            )
            features.extend(feats.values())

        return np.array(features)

    def citations(self):
        return ["Kevin Maik Jablonka and Berend Smit, TBA."]

    def implementors(self):
        return ["Kevin Maik Jablonka"]



# -*- coding: utf-8 -*-
"""Helper to build bench model for models that operate on pre-computed feature frames."""

from typing import Collection, Optional

from loguru import logger

__all__ = ["DFModel"]


class DFModel:
    def __init__(self, model, features: Optional[Collection[str]] = None):
        """Initialize the model.

        .. note:::

            If you use the model, you must set
            :code:`patch_in_ds` in the
            :code:py:`~mofdscribe.bench.mofbench.MOFBench` class.

        Args:
            model (object): Must implement `fit` and `predict` methods.
                Using a sklearn function signature will work.
            features (Collection[str], optional): Feature names to use.
                If not provided, all features will be used.
                Defaults to None.
        """
        self._model = model
        self._features = features

    @property
    def features(self):
        if self._features is None:
            self._features = list(self.ds.available_features)
        return self._features

    def fit(self, idx, structures, y):
        logger.debug("Fitting model")
        X = self.ds._df[self.features].loc[idx, :]  # noqa: N806
        logger.debug(X.shape)
        self._model.fit(X, y)  # noqa: N806

    def predict(self, idx, structures):
        X = self.ds._df[self.features].loc[idx, :]  # noqa: N806
        return self._model.predict(X)  # noqa: N806



# -*- coding: utf-8 -*-
"""Functions that can be used to aggregate values/lists of values"""

import numpy as np
from scipy.stats import gmean, hmean
from scipy.stats.mstats import gmean as mstast_gmean
from scipy.stats.mstats import hmean as mstast_hmean


def ma_percentile(values, percentile):
    mdata = np.ma.filled(values.astype(float), np.nan)
    return np.nanpercentile(mdata, percentile)


def trimean(values: np.typing.ArrayLike) -> float:
    r"""Calculate the trimean of the values:

    .. math::

        TM={\frac {Q_{1}+2Q_{2}+Q_{3}}{4}}

    Args:
        values (np.typing.ArrayLike): values to compute the trimean for

    Returns:
        float: trimean
    """
    q1 = np.percentile(values, 25)
    q2 = np.percentile(values, 50)
    q3 = np.percentile(values, 75)
    return (q1 + 2 * q2 + q3) / 4


def masked_mad(values):
    diff = values - np.ma.median(values)
    return np.ma.median(np.ma.abs(diff))


def mad(values):
    diff = values - np.median(values)
    return np.median(np.abs(diff))


def masked_trimean(values):
    q1 = ma_percentile(values, 25)
    q2 = ma_percentile(values, 50)
    q3 = ma_percentile(values, 75)
    return (q1 + 2 * q2 + q3) / 4


AGGREGATORS = {
    "sum": lambda x: x[0] + x[1],
    "avg": lambda x: (x[0] + x[1]) / 2,
    "max": lambda x: max(x),
    "min": lambda x: min(x),
    "first": lambda x: x[0],
    "last": lambda x: x[1],
    "product": lambda x: x[0] * x[1],
    "diff": lambda x: abs(x[1] - x[0]),
    "absdiff": lambda x: abs(x[1] - x[0]),
}


ARRAY_AGGREGATORS = {
    "sum": lambda x, **kwargs: np.sum(x, **kwargs),
    "avg": lambda x, **kwargs: np.mean(x, **kwargs),
    "max": lambda x, **kwargs: np.max(x, **kwargs),
    "min": lambda x, **kwargs: np.min(x, **kwargs),
    "std": lambda x, **kwargs: np.std(x, **kwargs),
    "range": lambda x, **kwargs: np.max(x, **kwargs) - np.min(x, **kwargs),
    "mean": lambda x, **kwargs: np.mean(x, **kwargs),
    "median": lambda x, **kwargs: np.median(x, **kwargs),
    "geomean": lambda x, **kwargs: gmean(x, **kwargs),
    "harmean": lambda x, **kwargs: hmean(x, **kwargs),
    "mad": lambda x, **kwargs: mad(x, **kwargs),
    "trimean": lambda x, **kwargs: trimean(x, **kwargs),
    "inf": lambda x, **kwargs: np.linalg.norm(x, ord=np.inf, **kwargs),
    "manhattan": lambda x, **kwargs: np.linalg.norm(x, ord=1, **kwargs),
}


MA_ARRAY_AGGREGATORS = {
    "sum": lambda x, **kwargs: np.ma.sum(x, **kwargs),
    "avg": lambda x, **kwargs: np.ma.mean(x, **kwargs),
    "max": lambda x, **kwargs: np.ma.max(x, **kwargs),
    "min": lambda x, **kwargs: np.ma.min(x, **kwargs),
    "std": lambda x, **kwargs: np.ma.std(x, **kwargs),
    "range": lambda x, **kwargs: np.ma.max(x, **kwargs) - np.ma.min(x, **kwargs),
    "mean": lambda x, **kwargs: np.ma.mean(x, **kwargs),
    "median": lambda x, **kwargs: np.ma.median(x, **kwargs),
    "geomean": lambda x, **kwargs: mstast_gmean(x, **kwargs),
    "harmean": lambda x, **kwargs: mstast_hmean(x, **kwargs),
    "mad": lambda x, **kwargs: masked_mad(x, **kwargs),
    "trimean": lambda x, **kwargs: masked_trimean(x, **kwargs),
    "inf": lambda x, **kwargs: np.linalg.norm(x, ord=np.inf, **kwargs),
    "manhattan": lambda x, **kwargs: np.linalg.norm(x, ord=1, **kwargs),
}



# -*- coding: utf-8 -*-
"""Describe molecules by computing statistics of pairwise distances between their atoms."""
from typing import List, Tuple, Union

import numpy as np
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from mofdscribe.featurizers.utils.aggregators import ARRAY_AGGREGATORS
from mofdscribe.featurizers.utils.extend import (
    operates_on_imolecule,
    operates_on_istructure,
    operates_on_molecule,
    operates_on_structure,
)


@operates_on_molecule
@operates_on_imolecule
@operates_on_istructure
@operates_on_structure
class PairwiseDistanceStats(BaseFeaturizer):
    """
    Describe the shape of molecules by computing statistics of pairwise distances.

    For doing so, we will just compute all pairwise distances and then compute
    some statistics on them.
    One might also think of this as pretty rough approximation of something like
    the AMD fingerpint.
    """

    def __init__(self, aggregations: Tuple[str] = ("mean", "std", "max", "min")) -> None:
        """Create a new PairwiseDistanceStats featurizer.

        Args:
            aggregations (Tuple[str], optional): Aggregations to compute over the pairwise
                distances. Must be one of :py:obj:`ARRAY_AGGREGATORS`.
                Defaults to ("mean", "std", "max", "min").
        """
        self.aggregations = aggregations

    def feature_labels(self) -> List[str]:
        return [f"pairwise_distance_stats_{a}" for a in self.aggregations]

    def featurize(self, structure: Union[Molecule, IMolecule, Structure, IStructure]) -> np.ndarray:
        distances = []
        for i, _ in enumerate(structure):
            for j, _ in enumerate(structure):
                if i < j:
                    distances.append(structure.get_distance(i, j))

        features = []
        for agg in self.aggregations:
            features.append(ARRAY_AGGREGATORS[agg](distances))
        return np.array(features)

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka"]

    def citations(self) -> List[str]:
        return []



# -*- coding: utf-8 -*-
"""Test the VoxelGrid featurizer."""
from mofdscribe.featurizers.pore.voxelgrid import VoxelGrid


def test_voxelgrid(hkust_structure):
    """Ensure we get the correct number of features."""
    vg = VoxelGrid()
    fl = vg.feature_labels()
    features = vg.featurize(hkust_structure)
    assert len(fl) == len(features)



# -*- coding: utf-8 -*-
"""Featurizers operating on the secondary building units."""
from .bu_featurizer import BUFeaturizer, MOFBBs  # noqa: F401
from .bu_matches import BUMatch  # noqa: F401
from .compositionstats_featurizer import CompositionStats  # noqa: F401
from .distance_hist_featurizer import PairwiseDistanceHist  # noqa: F401
from .distance_stats_featurizer import PairwiseDistanceStats  # noqa: F401
from .lsop_featurizer import LSOP  # noqa: F401
from .nconf20_featurizer import NConf20  # noqa: F401
from .rdkitadaptor import RDKitAdaptor  # noqa: F401
from .shape_featurizer import (  # noqa: F401
    PMI1,
    PMI2,
    PMI3,
    Asphericity,
    DiskLikeness,
    Eccentricity,
    InertialShapeFactor,
    RadiusOfGyration,
    RodLikeness,
    SphereLikeness,
    SpherocityIndex,
)



# -*- coding: utf-8 -*-
"""
Routines for the conversion between pymatgen molecules and RDKit.

Mostly copied from
https://github.com/mjwen/bondnet/blob/b719bd85235012c567298cb0da81ef113e3744bc/bondnet/core/rdmol.py#L265
which is licensed under CDDLv1.0.

COMMON DEVELOPMENT AND DISTRIBUTION LICENSE (CDDL) Version 1.0

1. Definitions.

1.1. "Contributor" means each individual or entity that
creates or contributes to the creation of Modifications.

1.2. "Contributor Version" means the combination of the
Original Software, prior Modifications used by a
Contributor (if any), and the Modifications made by that
particular Contributor.

1.3. "Covered Software" means (a) the Original Software, or
(b) Modifications, or (c) the combination of files
containing Original Software with files containing
Modifications, in each case including portions thereof.

1.4. "Executable" means the Covered Software in any form
other than Source Code.

1.5. "Initial Developer" means the individual or entity
that first makes Original Software available under this
License.

1.6. "Larger Work" means a work which combines Covered
Software or portions thereof with code not governed by the
terms of this License.

1.7. "License" means this document.

1.8. "Licensable" means having the right to grant, to the
maximum extent possible, whether at the time of the initial
grant or subsequently acquired, any and all of the rights
conveyed herein.

1.9. "Modifications" means the Source Code and Executable
form of any of the following:

A. Any file that results from an addition to,
deletion from or modification of the contents of a
file containing Original Software or previous
Modifications;

B. Any new file that contains any part of the
Original Software or previous Modification; or

C. Any new file that is contributed or otherwise made
available under the terms of this License.

1.10. "Original Software" means the Source Code and
Executable form of computer software code that is
originally released under this License.

1.11. "Patent Claims" means any patent claim(s), now owned
or hereafter acquired, including without limitation,
method, process, and apparatus claims, in any patent
Licensable by grantor.

1.12. "Source Code" means (a) the common form of computer
software code in which modifications are made and (b)
associated documentation included in or with such code.

1.13. "You" (or "Your") means an individual or a legal
entity exercising rights under, and complying with all of
the terms of, this License. For legal entities, "You"
includes any entity which controls, is controlled by, or is
under common control with You. For purposes of this
definition, "control" means (a) the power, direct or
indirect, to cause the direction or management of such
entity, whether by contract or otherwise, or (b) ownership
of more than fifty percent (50%) of the outstanding shares
or beneficial ownership of such entity.

2. License Grants.

2.1. The Initial Developer Grant.

Conditioned upon Your compliance with Section 3.1 below and
subject to third party intellectual property claims, the
Initial Developer hereby grants You a world-wide,
royalty-free, non-exclusive license:

(a) under intellectual property rights (other than
patent or trademark) Licensable by Initial Developer,
to use, reproduce, modify, display, perform,
sublicense and distribute the Original Software (or
portions thereof), with or without Modifications,
and/or as part of a Larger Work; and

(b) under Patent Claims infringed by the making,
using or selling of Original Software, to make, have
made, use, practice, sell, and offer for sale, and/or
otherwise dispose of the Original Software (or
portions thereof).

(c) The licenses granted in Sections 2.1(a) and (b)
are effective on the date Initial Developer first
distributes or otherwise makes the Original Software
available to a third party under the terms of this
License.

(d) Notwithstanding Section 2.1(b) above, no patent
license is granted: (1) for code that You delete from
the Original Software, or (2) for infringements
caused by: (i) the modification of the Original
Software, or (ii) the combination of the Original
Software with other software or devices.

2.2. Contributor Grant.

Conditioned upon Your compliance with Section 3.1 below and
subject to third party intellectual property claims, each
Contributor hereby grants You a world-wide, royalty-free,
non-exclusive license:

(a) under intellectual property rights (other than
patent or trademark) Licensable by Contributor to
use, reproduce, modify, display, perform, sublicense
and distribute the Modifications created by such
Contributor (or portions thereof), either on an
unmodified basis, with other Modifications, as
Covered Software and/or as part of a Larger Work; and

(b) under Patent Claims infringed by the making,
using, or selling of Modifications made by that
Contributor either alone and/or in combination with
its Contributor Version (or portions of such
combination), to make, use, sell, offer for sale,
have made, and/or otherwise dispose of: (1)
Modifications made by that Contributor (or portions
thereof); and (2) the combination of Modifications
made by that Contributor with its Contributor Version
(or portions of such combination).

(c) The licenses granted in Sections 2.2(a) and
2.2(b) are effective on the date Contributor first
distributes or otherwise makes the Modifications
available to a third party.

(d) Notwithstanding Section 2.2(b) above, no patent
license is granted: (1) for any code that Contributor
has deleted from the Contributor Version; (2) for
infringements caused by: (i) third party
modifications of Contributor Version, or (ii) the
combination of Modifications made by that Contributor
with other software (except as part of the
Contributor Version) or other devices; or (3) under
Patent Claims infringed by Covered Software in the
absence of Modifications made by that Contributor.

3. Distribution Obligations.

3.1. Availability of Source Code.

Any Covered Software that You distribute or otherwise make
available in Executable form must also be made available in
Source Code form and that Source Code form must be
distributed only under the terms of this License. You must
include a copy of this License with every copy of the
Source Code form of the Covered Software You distribute or
otherwise make available. You must inform recipients of any
such Covered Software in Executable form as to how they can
obtain such Covered Software in Source Code form in a
reasonable manner on or through a medium customarily used
for software exchange.

3.2. Modifications.

The Modifications that You create or to which You
contribute are governed by the terms of this License. You
represent that You believe Your Modifications are Your
original creation(s) and/or You have sufficient rights to
grant the rights conveyed by this License.

3.3. Required Notices.

You must include a notice in each of Your Modifications
that identifies You as the Contributor of the Modification.
You may not remove or alter any copyright, patent or
trademark notices contained within the Covered Software, or
any notices of licensing or any descriptive text giving
attribution to any Contributor or the Initial Developer.

3.4. Application of Additional Terms.

You may not offer or impose any terms on any Covered
Software in Source Code form that alters or restricts the
applicable version of this License or the recipients'
rights hereunder. You may choose to offer, and to charge a
fee for, warranty, support, indemnity or liability
obligations to one or more recipients of Covered Software.
However, you may do so only on Your own behalf, and not on
behalf of the Initial Developer or any Contributor. You
must make it absolutely clear that any such warranty,
support, indemnity or liability obligation is offered by
You alone, and You hereby agree to indemnify the Initial
Developer and every Contributor for any liability incurred
by the Initial Developer or such Contributor as a result of
warranty, support, indemnity or liability terms You offer.

3.5. Distribution of Executable Versions.

You may distribute the Executable form of the Covered
Software under the terms of this License or under the terms
of a license of Your choice, which may contain terms
different from this License, provided that You are in
compliance with the terms of this License and that the
license for the Executable form does not attempt to limit
or alter the recipient's rights in the Source Code form
from the rights set forth in this License. If You
distribute the Covered Software in Executable form under a
different license, You must make it absolutely clear that
any terms which differ from this License are offered by You
alone, not by the Initial Developer or Contributor. You
hereby agree to indemnify the Initial Developer and every
Contributor for any liability incurred by the Initial
Developer or such Contributor as a result of any such terms
You offer.

3.6. Larger Works.

You may create a Larger Work by combining Covered Software
with other code not governed by the terms of this License
and distribute the Larger Work as a single product. In such
a case, You must make sure the requirements of this License
are fulfilled for the Covered Software.

4. Versions of the License.

4.1. New Versions.

Sun Microsystems, Inc. is the initial license steward and
may publish revised and/or new versions of this License
from time to time. Each version will be given a
distinguishing version number. Except as provided in
Section 4.3, no one other than the license steward has the
right to modify this License.

4.2. Effect of New Versions.

You may always continue to use, distribute or otherwise
make the Covered Software available under the terms of the
version of the License under which You originally received
the Covered Software. If the Initial Developer includes a
notice in the Original Software prohibiting it from being
distributed or otherwise made available under any
subsequent version of the License, You must distribute and
make the Covered Software available under the terms of the
version of the License under which You originally received
the Covered Software. Otherwise, You may also choose to
use, distribute or otherwise make the Covered Software
available under the terms of any subsequent version of the
License published by the license steward.

4.3. Modified Versions.

When You are an Initial Developer and You want to create a
new license for Your Original Software, You may create and
use a modified version of this License if You: (a) rename
the license and remove any references to the name of the
license steward (except to note that the license differs
from this License); and (b) otherwise make it clear that
the license contains terms which differ from this License.

5. DISCLAIMER OF WARRANTY.

COVERED SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "AS IS"
BASIS, WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
INCLUDING, WITHOUT LIMITATION, WARRANTIES THAT THE COVERED
SOFTWARE IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR
PURPOSE OR NON-INFRINGING. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE COVERED SOFTWARE IS WITH YOU. SHOULD ANY
COVERED SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, YOU (NOT THE
INITIAL DEVELOPER OR ANY OTHER CONTRIBUTOR) ASSUME THE COST OF
ANY NECESSARY SERVICING, REPAIR OR CORRECTION. THIS DISCLAIMER OF
WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE. NO USE OF
ANY COVERED SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS
DISCLAIMER.

6. TERMINATION.

6.1. This License and the rights granted hereunder will
terminate automatically if You fail to comply with terms
herein and fail to cure such breach within 30 days of
becoming aware of the breach. Provisions which, by their
nature, must remain in effect beyond the termination of
this License shall survive.

6.2. If You assert a patent infringement claim (excluding
declaratory judgment actions) against Initial Developer or
a Contributor (the Initial Developer or Contributor against
whom You assert such claim is referred to as "Participant")
alleging that the Participant Software (meaning the
Contributor Version where the Participant is a Contributor
or the Original Software where the Participant is the
Initial Developer) directly or indirectly infringes any
patent, then any and all rights granted directly or
indirectly to You by such Participant, the Initial
Developer (if the Initial Developer is not the Participant)
and all Contributors under Sections 2.1 and/or 2.2 of this
License shall, upon 60 days notice from Participant
terminate prospectively and automatically at the expiration
of such 60 day notice period, unless if within such 60 day
period You withdraw Your claim with respect to the
Participant Software against such Participant either
unilaterally or pursuant to a written agreement with
Participant.

6.3. In the event of termination under Sections 6.1 or 6.2
above, all end user licenses that have been validly granted
by You or any distributor hereunder prior to termination
(excluding licenses granted to You by any distributor)
shall survive termination.

7. LIMITATION OF LIABILITY.

UNDER NO CIRCUMSTANCES AND UNDER NO LEGAL THEORY, WHETHER TORT
(INCLUDING NEGLIGENCE), CONTRACT, OR OTHERWISE, SHALL YOU, THE
INITIAL DEVELOPER, ANY OTHER CONTRIBUTOR, OR ANY DISTRIBUTOR OF
COVERED SOFTWARE, OR ANY SUPPLIER OF ANY OF SUCH PARTIES, BE
LIABLE TO ANY PERSON FOR ANY INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES OF ANY CHARACTER INCLUDING, WITHOUT
LIMITATION, DAMAGES FOR LOST PROFITS, LOSS OF GOODWILL, WORK
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER
COMMERCIAL DAMAGES OR LOSSES, EVEN IF SUCH PARTY SHALL HAVE BEEN
INFORMED OF THE POSSIBILITY OF SUCH DAMAGES. THIS LIMITATION OF
LIABILITY SHALL NOT APPLY TO LIABILITY FOR DEATH OR PERSONAL
INJURY RESULTING FROM SUCH PARTY'S NEGLIGENCE TO THE EXTENT
APPLICABLE LAW PROHIBITS SUCH LIMITATION. SOME JURISDICTIONS DO
NOT ALLOW THE EXCLUSION OR LIMITATION OF INCIDENTAL OR
CONSEQUENTIAL DAMAGES, SO THIS EXCLUSION AND LIMITATION MAY NOT
APPLY TO YOU.

8. U.S. GOVERNMENT END USERS.

The Covered Software is a "commercial item," as that term is
defined in 48 C.F.R. 2.101 (Oct. 1995), consisting of "commercial
computer software" (as that term is defined at 48 C.F.R.
252.227-7014(a)(1)) and "commercial computer software
documentation" as such terms are used in 48 C.F.R. 12.212 (Sept.
1995). Consistent with 48 C.F.R. 12.212 and 48 C.F.R. 227.7202-1
through 227.7202-4 (June 1995), all U.S. Government End Users
acquire Covered Software with only those rights set forth herein.
This U.S. Government Rights clause is in lieu of, and supersedes,
any other FAR, DFAR, or other clause or provision that addresses
Government rights in computer software under this License.

9. MISCELLANEOUS.

This License represents the complete agreement concerning subject
matter hereof. If any provision of this License is held to be
unenforceable, such provision shall be reformed only to the
extent necessary to make it enforceable. This License shall be
governed by the law of the jurisdiction specified in a notice
contained within the Original Software (except to the extent
applicable law, if any, provides otherwise), excluding such
jurisdiction's conflict-of-law provisions. Any litigation
relating to this License shall be subject to the jurisdiction of
the courts located in the jurisdiction and venue specified in a
notice contained within the Original Software, with the losing
party responsible for costs, including, without limitation, court
costs and reasonable attorneys' fees and expenses. The
application of the United Nations Convention on Contracts for the
International Sale of Goods is expressly excluded. Any law or
regulation which provides that the language of a contract shall
be construed against the drafter shall not apply to this License.
You agree that You alone are responsible for compliance with the
United States export administration regulations (and the export
control laws and regulation of any other countries) when You use,
distribute or otherwise make available any Covered Software.

10. RESPONSIBILITY FOR CLAIMS.

As between Initial Developer and the Contributors, each party is
responsible for claims and damages arising, directly or
indirectly, out of its utilization of rights under this License
and You agree to work with Initial Developer and Contributors to
distribute such responsibility on an equitable basis. Nothing
herein is intended or shall be deemed to constitute any admission
of liability.
"""
from collections import defaultdict
from ctypes import Structure
from typing import Collection, Dict, Optional, Tuple

import numpy as np
from loguru import logger
from numpy.typing import ArrayLike
from pymatgen.core import Molecule
from rdkit import Chem
from rdkit.Chem import BondType
from rdkit.Geometry import Point3D

METALS = {
    # first group
    "Li": 1,
    "Na": 1,
    "K": 1,
    "Rb": 1,
    "Cs": 1,
    # second group
    "Be": 2,
    "Mg": 2,
    "Ca": 2,
    "Sr": 2,
    "Ba": 2,
    "Al": 3,
    "Sc": 1,
    "Ti": 4,
    "V": 2,
    "Cr": 2,
    "Mn": 2,
    "Fe": 2,
    "Co": 2,
    "Ni": 2,
    "Cu": 2,
    "Zn": 2,
    "Zr": 4,
    # Lanthanides and Actinides
    "La": 3,
    "Ce": 3,
    "Pr": 3,
    "Nd": 3,
    "Pm": 3,
    "Sm": 3,
    "Eu": 3,
    "Gd": 3,
    "Tb": 3,
    "Dy": 3,
    "Ho": 3,
    "Er": 3,
    "Tm": 3,
    "Yb": 3,
    "Lu": 3,
    "Hf": 4,
    "Ta": 4,
    "U": 4,
    "W": 4,
    "Re": 4,
    "Os": 4,
    "Ir": 4,
    "Pt": 4,
}


def pymatgen_2_babel_atom_idx_map(pmg_mol: Molecule, ob_mol) -> Dict[int, int]:
    """
    Create an atom index mapping between pymatgen mol and openbabel mol.

    This does not require pymatgen mol and ob mol has the same number of atoms.
    But ob_mol can have smaller number of atoms.

    Args:
        pmg_mol (Molecule): pymatgen molecule
        ob_mol (ob.OBMol): openbabel molecule

    Returns:
        Dict[int, int]: with atom index in pymatgen mol as key and atom index in babel mol as
            value. Value is `None` if there is not corresponding atom in babel.

    Raises:
        RuntimeError: if mapping is not possible.
    """
    from openbabel import openbabel as ob

    pmg_coords = pmg_mol.cart_coords
    ob_coords = [[a.GetX(), a.GetY(), a.GetZ()] for a in ob.OBMolAtomIter(ob_mol)]
    ob_index = [a.GetIdx() for a in ob.OBMolAtomIter(ob_mol)]

    mapping = {i: None for i in range(len(pmg_coords))}

    for idx, oc in zip(ob_index, ob_coords):
        for i, gc in enumerate(pmg_coords):
            if np.allclose(oc, gc):
                mapping[i] = idx
                break
        else:
            raise RuntimeError("Cannot create atom index mapping pymatgen and ob mols")

    return mapping


def create_rdkit_mol(
    species: Collection[str],
    coords: ArrayLike,
    bond_types: Dict[Tuple[int, int], Chem.rdchem.BondType],
    formal_charge: Optional[Collection[float]] = None,
    name: Optional[str] = None,
    force_sanitize: bool = True,
):
    """
    Create a rdkit mol from scratch.

    Followed: https://sourceforge.net/p/rdkit/mailman/message/36474923/

    Args:
        species (Collection[str],): species str of each molecule
        coords (ArrayLike): positions of atoms
        bond_types (Dict[Tuple[int, int]): with bond indices (2 tuple) as key and bond type
            (e.g. Chem.rdchem.BondType.DOUBLE) as value
        formal_charge (Collection[float], optional): formal charge of each atom.
            Defaults to None.
        name (str, optional): name of the molecule.
            Defaults to None.
        force_sanitize (bool): whether to force the sanitization of molecule.
            If `True` and the sanitization fails, it generally throw an error
            and then stops. If `False`, will try to sanitize first, but if it fails,
            will proceed smoothly giving a warning message.

    Returns:
        rdkit Chem.Mol
    """
    m = Chem.Mol()
    edm = Chem.EditableMol(m)
    conformer = Chem.Conformer(len(species))

    for i, (s, c) in enumerate(zip(species, coords)):
        atom = Chem.Atom(s)
        atom.SetNoImplicit(True)
        if formal_charge is not None:
            cg = formal_charge[i]
            if cg is not None:
                atom.SetFormalCharge(cg)
        atom_idx = edm.AddAtom(atom)
        conformer.SetAtomPosition(atom_idx, Point3D(*c))

    for b, t in bond_types.items():
        edm.AddBond(b[0], b[1], t)

    m = edm.GetMol()
    if force_sanitize:
        Chem.SanitizeMol(m)
    else:
        try:
            Chem.SanitizeMol(m)
        except Exception as e:
            logger.warning(f"Cannot sanitize molecule {name}, because {str(e)}")
    m.AddConformer(conformer, assignId=False)

    m.SetProp("_Name", str(name))

    return m


def adjust_formal_charge(species: Collection[str], bonds: Collection[Tuple[int, int]]):
    """
    Adjust formal charge of metal atoms.

    Args:
        species (Collection[str]): species string of atoms
        bonds (Collection[Tuple[int, int]]): 2-tuple index of bonds

    Returns:
        list: formal charge of atoms. None for non metal atoms.

    ToDo:
        - use something like oximachine to guess oxidation state
    """
    # initialize formal charge first so that atom does not form any bond has its formal
    # charge set
    formal_charge = [METALS[s] if s in METALS else None for s in species]

    # atom_idx: idx of atoms in the molecule
    # num_bonds: number of bonds the atom forms
    atom_idx, num_bonds = np.unique(bonds, return_counts=True)
    for i, ct in zip(atom_idx, num_bonds):
        s = species[i]
        if s in METALS:
            formal_charge[i] = int(formal_charge[i] - ct)

    return formal_charge


def remove_metals(mol: Molecule):
    """
    Check whether metals are in a pymatgen molecule. If yes, create a new Molecule with metals removed.

    Args:
        mol (Molecule): pymatgen molecule

    Returns:
        pymatgen mol
    """
    species = [str(s) for s in mol.species]

    #  metals (dict): with metal specie are key and charge as value

    if set(species).intersection(set(METALS.keys())):
        charge = mol.charge

        species = []
        coords = []
        properties = defaultdict(list)
        for site in mol:
            s = str(site.specie)
            if s in METALS:
                charge -= METALS[s]
            else:
                species.append(s)
                coords.append(site.coords)
                for k, v in site.properties:
                    properties[k].append(v)

        # do not provide spin_multiplicity, since we remove an atom
        mol = Molecule(species, coords, charge, site_properties=properties)

    return mol


def create_rdkit_mol_from_mol_graph(
    mol_graph,
    name: Optional[str] = None,
    force_sanitize: bool = False,
) -> Chem.Mol:
    """
    Create a rdkit molecule from molecule graph, with bond type perceived by babel.

    Done in the below steps:
    1. create a babel mol without metal atoms.
    2. perceive bond order (conducted by BabelMolAdaptor)
    3. adjust formal charge of metal atoms so as not to violate valence rule
    4. create rdkit mol based on species, coords, bonds, and formal charge

    Args:
        mol_graph (pymatgen MoleculeGraph): molecule graph
        name (str, optional): name of the molecule.
            Defaults to None.
        force_sanitize (bool): whether to force sanitization of the rdkit mol

    Returns:
        mol (Chem.Mol): rdkit Chem.Mol

    Raises:
        RuntimeError: if it finds and unexpected bond type or a bond between
            two metals
    """
    from openbabel import openbabel as ob
    from pymatgen.io.babel import BabelMolAdaptor

    pymatgen_mol = mol_graph.molecule
    species = [str(s) for s in pymatgen_mol.species]
    coords = pymatgen_mol.cart_coords
    bonds = [tuple(sorted([i, j])) for i, j, _ in mol_graph.graph.edges.data()]

    # create babel mol without metals
    pmg_mol_no_metals = remove_metals(pymatgen_mol)
    adaptor = BabelMolAdaptor(pmg_mol_no_metals)
    ob_mol = adaptor.openbabel_mol

    # get babel bond order of mol without metals
    ob_bond_order = {}
    for bd in ob.OBMolBondIter(ob_mol):
        k = tuple(sorted([bd.GetBeginAtomIdx(), bd.GetEndAtomIdx()]))
        v = bd.GetBondOrder()
        ob_bond_order[k] = v

    # create bond type
    atom_idx_mapping = pymatgen_2_babel_atom_idx_map(pymatgen_mol, ob_mol)
    bond_types = {}

    for bd in bonds:
        try:
            ob_bond = [atom_idx_mapping[a] for a in bd]

            # atom not in ob mol
            if None in ob_bond:
                raise KeyError
            # atom in ob mol
            else:
                ob_bond = tuple(sorted(ob_bond))
                v = ob_bond_order[ob_bond]
                if v == 0:
                    tp = BondType.UNSPECIFIED
                elif v == 1:
                    tp = BondType.SINGLE
                elif v == 2:
                    tp = BondType.DOUBLE
                elif v == 3:
                    tp = BondType.TRIPLE
                elif v == 5:
                    tp = BondType.AROMATIC
                else:
                    raise RuntimeError(f"Got unexpected babel bond order: {v}")

        except KeyError:
            atom1_spec, atom2_spec = [species[a] for a in bd]

            if atom1_spec in METALS and atom2_spec in METALS:
                raise RuntimeError("Got a bond between two metal atoms")

            # bond involves one and only one metal atom (atom not in ob mol case above)
            elif atom1_spec in METALS or atom2_spec in METALS:
                tp = Chem.rdchem.BondType.DATIVE

                # Dative bonds have the special characteristic that they do not affect
                # the valence on the start atom, but do affect the end atom.
                # Here we adjust the atom ordering in the bond for dative bond to make
                # metal the end atom.
                if atom1_spec in METALS:
                    bd = tuple(reversed(bd))

            # bond not found by babel (atom in ob mol)
            else:
                tp = Chem.rdchem.BondType.UNSPECIFIED

        bond_types[bd] = tp

    # a metal atom can form multiple dative bond (e.g. bidentate LiEC), for such cases
    # we need to adjust the their formal charge so as not to violate valence rule
    formal_charge = adjust_formal_charge(species, bonds)

    m = create_rdkit_mol(species, coords, bond_types, formal_charge, name, force_sanitize)

    return m


def boxed_molecule(molecule: Molecule) -> Structure:
    """
    Create a box molecule from a pymatgen molecule.

    Args:
        molecule (Molecule): pymatgen molecule

    Returns:
        structure (Structure): molecule in a box
    """
    max_dist = molecule.distance_matrix.max()

    return molecule.get_boxed_structure(a=max_dist, b=max_dist, c=max_dist)



# -*- coding: utf-8 -*-
"""Data objects for holding collections of metrics (and performing basic aggregations)."""
from typing import List, Optional

import numpy as np
from pydantic import BaseModel

from .regression import RegressionMetrics, RegressionMetricsConcat


class RegressionMetricCollection(BaseModel):
    """Model for regression metric collection."""

    regression_metrics: List[RegressionMetrics]
    fit_timings: Optional[List[float]]
    inference_timings: Optional[List[float]]

    def concatenated_metrics(self) -> RegressionMetrics:
        """Return concatenated metrics."""
        return RegressionMetricsConcat(
            mean_squared_error=[metric.mean_squared_error for metric in self.regression_metrics],
            mean_absolute_error=[metric.mean_absolute_error for metric in self.regression_metrics],
            r2_score=[metric.r2_score for metric in self.regression_metrics],
            max_error=[metric.max_error for metric in self.regression_metrics],
            mean_absolute_percentage_error=[
                metrics.mean_absolute_percentage_error for metrics in self.regression_metrics
            ],
            top_100_in_top_100=[metric.top_100_in_top_100 for metric in self.regression_metrics],
            top_500_in_top_500=[metric.top_500_in_top_500 for metric in self.regression_metrics],
            top_50_in_top_50=[metric.top_50_in_top_50 for metric in self.regression_metrics],
            top_10_in_top_10=[metric.top_10_in_top_10 for metric in self.regression_metrics],
            top_5_in_top_5=[metric.top_5_in_top_5 for metric in self.regression_metrics],
        )

    def average_metrics(self) -> RegressionMetrics:
        """Return mean of metrics using numpy"""
        return RegressionMetrics(
            mean_squared_error=np.mean(
                [metric.mean_squared_error for metric in self.regression_metrics]
            ),
            mean_absolute_error=np.mean(
                [metric.mean_absolute_error for metric in self.regression_metrics]
            ),
            r2_score=np.mean([metric.r2_score for metric in self.regression_metrics]),
            max_error=np.mean([metric.max_error for metric in self.regression_metrics]),
            mean_absolute_percentage_error=np.mean(
                [metric.mean_absolute_percentage_error for metric in self.regression_metrics]
            ),
            top_100_in_top_100=np.mean(
                [metric.top_100_in_top_100 for metric in self.regression_metrics]
            ),
            top_500_in_top_500=np.mean(
                [metric.top_500_in_top_500 for metric in self.regression_metrics]
            ),
            top_50_in_top_50=np.mean(
                [metric.top_50_in_top_50 for metric in self.regression_metrics]
            ),
            top_10_in_top_10=np.mean(
                [metric.top_10_in_top_10 for metric in self.regression_metrics]
            ),
            top_5_in_top_5=np.mean([metric.top_5_in_top_5 for metric in self.regression_metrics]),
        )

    def std_metrics(self) -> RegressionMetrics:
        """Compute standard deviation of metrics using numpy."""
        return RegressionMetrics(
            mean_squared_error=np.std(
                [metric.mean_squared_error for metric in self.regression_metrics]
            ),
            mean_absolute_error=np.std(
                [metric.mean_absolute_error for metric in self.regression_metrics]
            ),
            r2_score=np.std([metric.r2_score for metric in self.regression_metrics]),
            max_error=np.std([metric.max_error for metric in self.regression_metrics]),
            mean_absolute_percentage_error=np.std(
                [metric.mean_absolute_percentage_error for metric in self.regression_metrics]
            ),
            top_100_in_top_100=np.std(
                [metric.top_100_in_top_100 for metric in self.regression_metrics]
            ),
            top_500_in_top_500=np.std(
                [metric.top_500_in_top_500 for metric in self.regression_metrics]
            ),
            top_50_in_top_50=np.std(
                [metric.top_50_in_top_50 for metric in self.regression_metrics]
            ),
            top_10_in_top_10=np.std(
                [metric.top_10_in_top_10 for metric in self.regression_metrics]
            ),
            top_5_in_top_5=np.std([metric.top_5_in_top_5 for metric in self.regression_metrics]),
        )

    def average_fit_timings(self) -> float:
        """Return mean of fit timings using numpy."""
        return np.mean(self.fit_timings)

    def std_fit_timings(self) -> float:
        """Return standard deviation of fit timings using numpy."""
        return np.std(self.fit_timings)

    def average_inferences_timings(self) -> float:
        """Return mean of inference timings using numpy."""
        return np.mean(self.inference_timings)

    def std_inference_timings(self) -> float:
        """Return standard deviation of inference timings using numpy."""
        return np.std(self.inference_timings)



# -*- coding: utf-8 -*-
"""Interface for creating a custom StructureDataset."""
from pathlib import Path
from typing import Collection, List, Optional

import numpy as np
import pandas as pd
from loguru import logger

from mofdscribe.datasets.checks import check_all_file_exists
from mofdscribe.datasets.dataset import AbstractStructureDataset
from mofdscribe.datasets.utils import compress_dataset
from mofdscribe.types import PathType

__all__ = ["StructureDataset", "FrameDataset"]


class StructureDataset(AbstractStructureDataset):
    """Custom dataset class for loading structures from a files"""

    def __init__(
        self,
        files: Collection[PathType],
        df: Optional[pd.DataFrame] = None,
        structure_name_column: Optional[str] = None,
        year_column: Optional[str] = None,
        label_columns: Optional[List[str]] = None,
        feature_columns: Optional[List[str]] = None,
        decorated_graph_hash_column: Optional[str] = None,
        undecorated_graph_hash_column: Optional[str] = None,
        decorated_scaffold_hash_column: Optional[str] = None,
        undecorated_scaffold_hash_column: Optional[str] = None,
        density_column: Optional[str] = None,
    ):
        """Initialize a dataset.

        Args:
            files (Collection[PathType]): List of files to load structures from.
            df (Optional[pd.DataFrame], optional): Dataframe containing the structures.
                Defaults to None.
            structure_name_column (str): Name of the column containing the structure names.
                Defaults to None.
            year_column (str, optional): Name of the column containing the year of the structure.
                Defaults to None.
            label_columns (Optional[List[str]], optional): List of columns containing the labels.
                Defaults to None.
            feature_columns (Optional[List[str]], optional): List of columns containing the features.
                Defaults to None.
            decorated_graph_hash_column (str, optional): Name of the column containing the decorated graph hash.
                Defaults to None.
            undecorated_graph_hash_column (str, optional): Name of the column containing the undecorated graph hash.
                Defaults to None.
            decorated_scaffold_hash_column (str, optional): Name of the column containing the decorated scaffold hash.
                Defaults to None.
            undecorated_scaffold_hash_column (str, optional): Name of the column containing
                the undecorated scaffold hash.
                Defaults to None.
            density_column (str, optional): Name of the column containing the density of the structure.
                Defaults to None.
        """
        super().__init__()
        self._df = df
        if self._df is not None:
            compress_dataset(self._df)
            self._structures = [
                f for f in files if Path(f).stem in self._df[structure_name_column].values
            ]
        else:
            self._structures = files
        self._structure_name_column = structure_name_column
        check_all_file_exists(self._structures)

        self._year_column = year_column
        self._label_columns = list(label_columns) if label_columns is not None else ()
        self._feature_columns = list(feature_columns) if feature_columns is not None else ()
        self._decorated_graph_hash_column = decorated_graph_hash_column
        self._undecorated_graph_hash_column = undecorated_graph_hash_column
        self._decorated_scaffold_hash_column = decorated_scaffold_hash_column
        self._undecorated_scaffold_hash_column = undecorated_scaffold_hash_column
        self._density_column = density_column

        self._years = None if year_column is None else self._df[year_column].values
        self._labels = None if label_columns is None else self._df[label_columns].values
        self._decorated_graph_hashes = (
            None
            if decorated_graph_hash_column is None
            else self._df[decorated_graph_hash_column].values
        )
        self._undecorated_graph_hashes = (
            None
            if undecorated_graph_hash_column is None
            else self._df[undecorated_graph_hash_column].values
        )
        self._decorated_scaffold_hashes = (
            None
            if decorated_scaffold_hash_column is None
            else self._df[decorated_scaffold_hash_column].values
        )
        self._undecorated_scaffold_hashes = (
            None
            if undecorated_scaffold_hash_column is None
            else self._df[undecorated_scaffold_hash_column].values
        )
        self._densities = None if density_column is None else self._df[density_column].values

    def __len__(self):
        """Return number of structures in the dataset."""
        return len(self._structures)

    @property
    def available_features(self) -> List[str]:
        return self._featurenames

    @property
    def available_labels(self) -> List[str]:
        return self._labelnames

    def get_labels(self, idx: Collection[int], labelnames: Collection[str] = None) -> np.ndarray:
        labelnames = labelnames if labelnames is not None else self._labelnames
        return self._df.iloc[idx][list(labelnames)].values

    @classmethod
    def from_folder_and_dataframe(
        cls,
        folder: PathType,
        extension: str = "cif",
        dataframe: Optional[pd.DataFrame] = None,
        structure_name_column: Optional[str] = None,
        year_column: Optional[str] = None,
        label_columns: Optional[List[str]] = None,
        decorated_graph_hash_column: Optional[str] = None,
        undecorated_graph_hash_column: Optional[str] = None,
        decorated_scaffold_hash_column: Optional[str] = None,
        undecorated_scaffold_hash_column: Optional[str] = None,
        density_column: Optional[str] = None,
    ) -> "StructureDataset":
        """Create a dataset from a folder and a dataframe.

        Args:
            folder (PathType): Path to the folder containing the structures.
            extension (str): Extension of the files. Defaults to 'cif'.
            dataframe (Optional[pd.DataFrame], optional): Dataframe containing the structures.
                Defaults to None.
            structure_name_column (str): Name of the column containing the structure names.
                Defaults to None.
            year_column (str, optional): Name of the column containing the year of the structure.
                Defaults to None.
            label_columns (Optional[List[str]], optional): List of columns containing the labels.
                Defaults to None.
            decorated_graph_hash_column (str, optional): Name of the column containing the decorated graph hash.
                Defaults to None.
            undecorated_graph_hash_column (str, optional): Name of the column containing the undecorated graph hash.
                Defaults to None.
            decorated_scaffold_hash_column (str, optional): Name of the column containing the decorated scaffold hash.
                Defaults to None.
            undecorated_scaffold_hash_column (str, optional): Name of the column containing the undecorated scaffold
                hash. Defaults to None.
            density_column (str, optional): Name of the column containing the density of the structure.
                Defaults to None.

        Returns:
            StructureDataset: Dataset containing the structures.
        """
        all_files = list(Path(folder).rglob(f"*.{extension}"))
        return cls(
            all_files,
            dataframe,
            structure_name_column,
            year_column,
            label_columns,
            decorated_graph_hash_column,
            undecorated_graph_hash_column,
            decorated_scaffold_hash_column,
            undecorated_scaffold_hash_column,
            density_column,
        )


class FrameDataset(AbstractStructureDataset):
    """Dataset containing structure information read from a dataframe."""

    def __init__(
        self,
        df: pd.DataFrame,
        structure_name_column: str,
        year_column: Optional[str] = None,
        label_columns: Optional[List[str]] = None,
        decorated_graph_hash_column: Optional[str] = None,
        undecorated_graph_hash_column: Optional[str] = None,
        decorated_scaffold_hash_column: Optional[str] = None,
        undecorated_scaffold_hash_column: Optional[str] = None,
        density_column: Optional[str] = None,
    ):
        """Initialize the dataset.

        Args:
            df (pd.DataFrame): Dataframe containing the structures.
            structure_name_column (str): Name of the column containing the structure names.
            year_column (str, optional): Name of the column containing the year of the structure.
                Defaults to None.
            label_columns (Optional[List[str]], optional): List of columns containing the labels.
                Defaults to None.
            decorated_graph_hash_column (str, optional): Name of the column containing the decorated graph hash.
                Defaults to None.
            undecorated_graph_hash_column (str, optional): Name of the column containing the undecorated graph hash.
                Defaults to None.
            decorated_scaffold_hash_column (str, optional): Name of the column containing the decorated scaffold hash.
                Defaults to None.
            undecorated_scaffold_hash_column (str, optional): Name of the column containing the undecorated scaffold
                hash. Defaults to None.
            density_column (str, optional): Name of the column containing the density of the structure.
                Defaults to None.
        """
        super().__init__()
        logger.warning(
            "FrameDataset support is experimental. Some splitter integrations may not work."
        )
        self._df = df
        compress_dataset(self._df)
        self._structure_name_column = structure_name_column
        self._year_column = year_column
        self._label_columns = list(label_columns) if label_columns is not None else ()
        self._decorated_graph_hash_column = decorated_graph_hash_column
        self._undecorated_graph_hash_column = undecorated_graph_hash_column
        self._decorated_scaffold_hash_column = decorated_scaffold_hash_column
        self._undecorated_scaffold_hash_column = undecorated_scaffold_hash_column
        self._density_column = density_column

        self._years = None if year_column is None else self._df[year_column]
        self._labels = None if label_columns is None else self._df[label_columns].values
        self._decorated_graph_hashes = (
            None
            if decorated_graph_hash_column is None
            else self._df[decorated_graph_hash_column].values
        )
        self._undecorated_graph_hashes = (
            None
            if undecorated_graph_hash_column is None
            else self._df[undecorated_graph_hash_column].values
        )
        self._decorated_scaffold_hashes = (
            None
            if decorated_scaffold_hash_column is None
            else self._df[decorated_scaffold_hash_column].values
        )
        self._undecorated_scaffold_hashes = (
            None
            if undecorated_scaffold_hash_column is None
            else self._df[undecorated_scaffold_hash_column].values
        )
        self._densities = None if density_column is None else self._df[density_column].values

    def __len__(self):
        """Return number of structures in the dataset."""
        return len(self._df)

    def get_subset(self, indices: Collection[int]) -> "FrameDataset":
        """Get a subset of the dataset.

        Args:
            indices (Collection[int]): indices of the structures to include.

        Returns:
            FrameDataset: a new dataset containing only the structures
                specified by the indices.
        """
        return FrameDataset(
            self._df.iloc[indices],
            structure_name_column=self._structure_name_column,
            year_column=self._year_column,
            label_columns=self._label_columns,
            decorated_graph_hash_column=self._decorated_graph_hash_column,
            undecorated_graph_hash_column=self._undecorated_graph_hash_column,
            decorated_scaffold_hash_column=self._decorated_scaffold_hash_column,
            undecorated_scaffold_hash_column=self._undecorated_scaffold_hash_column,
            density_column=self._density_column,
        )

    @property
    def available_features(self) -> List[str]:
        return self._featurenames

    @property
    def available_labels(self) -> List[str]:
        return self._labelnames

    def get_labels(self, idx: Collection[int], labelnames: Collection[str] = None) -> np.ndarray:
        labelnames = labelnames if labelnames is not None else self._labelnames
        return self._df.iloc[idx][list(labelnames)].values
